from ..functions import deepcopy
from .container import Container


class Properties(Container):
    """Mixin class for an object with descriptive properties.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, properties=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                   ``properties={'standard_name': 'altitude'}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                properties = source.properties()
            except AttributeError:
                properties = None

        self._set_component("properties", {}, copy=False)

        if properties is not None:
            self.set_properties(properties, copy=copy)

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def clear_properties(self):
        """Remove all properties.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_properties`, `del_property`, `properties`,
                     `set_properties`

        :Returns:

            `dict`
                The properties that have been removed.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.properties()
        {}
        >>> f.set_properties({'standard_name': 'air_pressure',
        ...                   'long_name': 'Air Pressure'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure'}
        >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.clear_properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.properties()
        {}

        """
        out = self._get_component("properties")
        self._set_component("properties", {}, copy=False)
        return out

    def del_property(self, prop, default=ValueError()):
        """Remove a property.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_properties`, `get_property`, `has_property`,
                     `properties`, `set_property`

        :Parameters:

            prop: `str`
                The name of the property to be removed.

                *Parameter example:*
                   ``prop='long_name'``

            default: optional
                Return the value of the *default* parameter if the
                property has not been set.

                {{default Exception}}

        :Returns:

                The removed property value.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        try:
            return self._get_component("properties").pop(prop)
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {prop!r} property",
            )

    def del_properties(self, properties):
        """Remove properties.

        .. versionadded:: (cfdm) 1.10.0.3

        .. seealso:: `clear_properties`, `del_property`, `properties`,
                     `set_properties`

        :Parameters:

            properties: (sequence of) `str`
                The names of the properties to be removed. If the
                `{{class}}` does not have a particular property then
                that property is ignored. No properties are removed if
                *properties* is an empty sequence.

                *Parameter example:*
                  ``'long_name'``

                *Parameter example:*
                  ``['long_name', 'comment']``

        :Returns:

            `dict`
                The removed property values, keyed by property name.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_properties({'project': 'CMIP7', 'comment': 'model'})
        >>> removed_properties = f.del_properties('project')
        >>> removed_properties
        {'project': 'CMIP7'}
        >>> f.properties()
        {'comment': 'model'}
        >>> f.set_properties(removed_properties)
        >>> f.properties()
        {'comment': 'model', 'project': 'CMIP7'}
        >>> f.del_properties('foo')
        {}

        """
        if isinstance(properties, str):
            properties = (properties,)

        out = {}
        for prop in properties:
            try:
                out[prop] = self.del_property(prop)
            except ValueError:
                pass

        return out

    def get_property(self, prop, default=ValueError()):
        """Return a property.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_property`, `has_property`, `properties`,
                     `set_property`

        :Parameters:

            prop: `str`
                The name of the property to be returned.

                *Parameter example:*
                   ``prop='standard_name'``

            default: optional
                Return the value of the *default* parameter if the
                property has not been set.

                {{default Exception}}

        :Returns:

                The value of the property.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        try:
            return self._get_component("properties")[prop]
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {prop!r} property",
            )

    def has_bounds(self):
        """Whether or not there are cell bounds.

        This is always False.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `has_data`

        :Returns:

            `False`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.has_bounds()
        False

        """
        return False

    def has_data(self):
        """Whether or not the construct has data.

        `{{class}}` instances never have data.

        .. versionadded:: (cfdm) 1.9.0.0

        :Returns:

            `False`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.has_data()
        False

        """
        return False

    def has_property(self, prop):
        """Whether a property has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_property`, `get_property`, `properties`,
                     `set_property`

        :Parameters:

            prop: `str`
                The name of the property.

                *Parameter example:*
                   ``prop='long_name'``

        :Returns:

            `bool`
                True if the property has been set, otherwise False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        return prop in self._get_component("properties")

    def properties(self):
        """Return all properties.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_properties`, `get_property`, `has_property`
                     `set_properties`

        :Returns:

            `dict`
                The properties.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.properties()
        {}
        >>> f.set_properties({'standard_name': 'air_pressure',
        ...                   'long_name': 'Air Pressure'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure'}
        >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.clear_properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.properties()
        {}

        """
        return self._get_component("properties").copy()

    def set_properties(self, properties, copy=True):
        """Set properties.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_properties`, `properties`, `set_property`

        :Parameters:

            properties: `dict`
                Store the properties from the dictionary supplied.

                *Parameter example:*
                  ``properties={'standard_name': 'altitude', 'foo': 'bar'}``

            copy: `bool`, optional
                If False then any property values provided by the
                *properties* parameter are not copied before insertion. By
                default they are deep copied.

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.properties()
        {}
        >>> f.set_properties({'standard_name': 'air_pressure',
        ...                   'long_name': 'Air Pressure'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure'}
        >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
        >>> f.properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.clear_properties()
        {'standard_name': 'air_pressure',
         'long_name': 'Air Pressure',
         'foo': 'bar'}
        >>> f.properties()
        {}

        """
        if copy:
            properties = deepcopy(properties)
        #        else:
        #            properties = properties.copy()

        self._get_component("properties").update(properties)

    def set_property(self, prop, value, copy=True):
        """Set a property.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_property`, `get_property`, `has_property`,
                     `properties`, `set_properties`

        :Parameters:

            prop: `str`
                The name of the property to be set.

            value:
                The value for the property.

            copy: `bool`, optional
                If True then set a deep copy of *value*.

        :Returns:

             `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_property('project', 'CMIP7')
        >>> f.has_property('project')
        True
        >>> f.get_property('project')
        'CMIP7'
        >>> f.del_property('project')
        'CMIP7'
        >>> f.has_property('project')
        False
        >>> print(f.del_property('project', None))
        None
        >>> print(f.get_property('project', None))
        None

        """
        if copy:
            value = deepcopy(value)

        self._get_component("properties")[prop] = value



import datetime
import faulthandler
import unittest

import numpy as np
from scipy.sparse import coo_array

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

row = np.array([0, 3, 1, 0])
col = np.array([0, 3, 1, 2])
data = np.array([4, 5, 7, 9])
s = coo_array((data, (row, col)), shape=(4, 4))


class SparseArrayTest(unittest.TestCase):
    """Unit test for the SparseArray class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows: cfdm.LOG_LEVEL('DEBUG')
        #
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_SparseArray_copy(self):
        """Test the copy module copying behaviour of SparseArray."""
        x = cfdm.SparseArray(s)
        self.assertTrue((s.toarray() == x.array).all())

    def test_SparseArray__array__(self):
        """Test the numpy array conversion of SparseArray."""
        x = cfdm.SparseArray(s)
        self.assertTrue((np.array(x) == x.array).all())

    def test_SparseArray_get_filename(self):
        """Test SparseArray.get_filename."""
        x = cfdm.SparseArray()
        with self.assertRaises(AttributeError):
            x.get_filename()

    def test_SparseArray_get_filenames(self):
        """Test SparseArray.get_filenames."""
        x = cfdm.SparseArray()
        with self.assertRaises(AttributeError):
            x.get_filenames()


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import numpy as np

from .quadraticgeographicinterpolation import QuadraticGeographicInterpolation


class QuadraticLatitudeLongitudeInterpolation(
    QuadraticGeographicInterpolation
):
    """Mixin class for quadratic latitude longitude interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _quadratic_latitude_longitude_interpolation(
        self, lat_a, lon_a, lat_b, lon_b, ce, ca, location_use_3d_cartesian, d1
    ):
        """Quadratic interpolation of geographic coordinates.

        A one-dimensional quadratic method for interpolation of the
        geographic coordinates latitude and longitude, typically used
        for remote sensing products with geographic coordinates on the
        reference ellipsoid.

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            lat_a, lon_a: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location A, in the sense of CF appendix J Figure J.3.

            lat_b, lon_b: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location B, in the sense of CF appendix J Figure J.3.

            ce, ca: `numpy.ndarray` or `None`
                The interpolation parameters ``ce`` and ``ca``, with
                the same number of dimensions in the same relative
                order as the tie points array. If any are `None` then
                the parameter is assumed to be zero.

            {{location_use_3d_cartesian: `numpy.ndarray` or `None`}}

            {{d1: `int`}}

        :Returns:

            `numpy.ndarray`
                The result of interpolating the tie points to interior
                locations implied by *s*.

        """
        if location_use_3d_cartesian is None:
            raise ValueError(
                "Can't uncompress tie points by "
                "quadratic_latitude_longitude interpolation "
                "when there is no interpolation_subarea_flags "
                "interpolation parameter that includes "
                "'location_use_3d_cartesian' in its flag_meanings "
                "attribute"
            )

        any_cartesian = bool(location_use_3d_cartesian.any())
        all_cartesian = bool(location_use_3d_cartesian.all())

        # Find out if these tie points are latitudes or longitudes. If
        # they are latitudes then the dependent tie points will be
        # longitudes.
        latitude = "longitude" in self.dependent_tie_points

        if latitude:
            fv2ll = self._fv2lat
        else:
            fv2ll = self._fv2lon

        va = self._fll2v(lat_a, lon_a)
        vb = self._fll2v(lat_b, lon_b)

        cv = self._fcea2cv(va, vb, ce, ca)

        if any_cartesian:
            # Interpolation in three-dimensional cartesian coordinates
            vp = self._fqv(va, vb, cv, d1)
            if all_cartesian:
                del va, vb, cv

            u_c = fv2ll(vp)
            del vp

            if all_cartesian:
                llp = u_c

        if not all_cartesian:
            # Interpolation in latitude-longitude coordinates
            if latitude:
                lla, llb = lat_a, lat_b
            else:
                lla, llb = lon_a, lon_b

            llab = fv2ll(self._fqv(va, vb, cv, 0.5))
            del va, vb, cv

            cll = self._fw(lla, llab, llab, s_i=0.5)
            del llab

            u_l = self._fq(lla, llb, cll, d1)
            del lla, llb, cll
            if latitude:
                del lat_a, lat_b
            else:
                del lon_a, lon_b

            if not any_cartesian:
                llp = u_l

        if any_cartesian and not all_cartesian:
            # Combine the results of cartesian and latitude-longitude
            # interpolations
            llp = np.where(location_use_3d_cartesian, u_c, u_l)

        return llp



import datetime
import faulthandler
import inspect
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


def _recurse_on_subclasses(klass):
    """Return all unique subclasses of a classes' subclass hierarchy."""
    return set(klass.__subclasses__()).union(
        [
            sub
            for cls in klass.__subclasses__()
            for sub in _recurse_on_subclasses(cls)
        ]
    )


def _get_all_abbrev_subclasses(klass):
    """Return subclasses of the class hierarchy with some filtered out.

    Filter out cf.mixin.properties*.Properties* (by means of there not
    being any abbreviated cf.Properties* classes) plus any cfdm classes,
    since this function needs to take cf subclasses from cfdm classes as
    well.

    """
    return tuple(
        [
            sub
            for sub in _recurse_on_subclasses(klass)
            if hasattr(cfdm, sub.__name__)
        ]
    )


class DocstringTest(unittest.TestCase):
    """Test the system for cross-package docstring substitutions."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        self.package = "cfdm"
        self.repr = ""

        self.subclasses_of_Container = tuple(
            set(
                _get_all_abbrev_subclasses(cfdm.mixin.container.Container)
            ).union(
                set(
                    _get_all_abbrev_subclasses(
                        cfdm.core.abstract.container.Container
                    )
                ),
                set(
                    _get_all_abbrev_subclasses(cfdm.data.abstract.array.Array)
                ),
            )
        )

        self.subclasses_of_Properties = _get_all_abbrev_subclasses(
            cfdm.mixin.properties.Properties
        )

        self.subclasses_of_PropertiesData = _get_all_abbrev_subclasses(
            cfdm.mixin.propertiesdata.PropertiesData
        )

        self.subclasses_of_PropertiesDataBounds = _get_all_abbrev_subclasses(
            cfdm.mixin.propertiesdatabounds.PropertiesDataBounds
        )

    def test_class_docstring_rewrite(self):
        """Test the DocstringRewriteMeta (meta)class."""

        class parent(metaclass=cfdm.core.meta.DocstringRewriteMeta):
            pass

        class child(parent):
            pass

        class grandchild(child):
            pass

        self.assertIsNone(parent.__doc__)
        self.assertIsNone(child.__doc__)
        self.assertIsNone(grandchild.__doc__)

        class parent(metaclass=cfdm.core.meta.DocstringRewriteMeta):
            """No sub 0."""

        class child(parent):
            pass

        class grandchild(child):
            pass

        self.assertEqual(parent.__doc__, "No sub 0.")
        self.assertIsNone(child.__doc__)
        self.assertIsNone(grandchild.__doc__)

        class parent(metaclass=cfdm.core.meta.DocstringRewriteMeta):
            """{{class}}."""

        class child(parent):
            pass

        class grandchild(child):
            pass

        self.assertEqual(parent.__doc__, "parent.")
        self.assertEqual(child.__doc__, "child.")
        self.assertEqual(grandchild.__doc__, "grandchild.")

        class child(parent):
            """No sub 1."""

        class grandchild(child):
            pass

        class greatgrandchild(grandchild):
            pass

        self.assertEqual(parent.__doc__, "parent.")
        self.assertEqual(child.__doc__, "No sub 1.")
        self.assertIsNone(grandchild.__doc__)
        self.assertIsNone(greatgrandchild.__doc__)

        class greatgrandchild(grandchild):
            """No sub 3."""

        self.assertEqual(parent.__doc__, "parent.")
        self.assertEqual(child.__doc__, "No sub 1.")
        self.assertIsNone(grandchild.__doc__)
        self.assertEqual(greatgrandchild.__doc__, "No sub 3.")

        class grandchild(child):
            """No sub 2."""

        class greatgrandchild(grandchild):
            pass

        self.assertEqual(parent.__doc__, "parent.")
        self.assertEqual(child.__doc__, "No sub 1.")
        self.assertEqual(grandchild.__doc__, "No sub 2.")
        self.assertIsNone(greatgrandchild.__doc__)

        class grandchild(child):
            """Sub 2 {{class}}."""

        class greatgrandchild(grandchild):
            pass

        self.assertEqual(parent.__doc__, "parent.")
        self.assertEqual(child.__doc__, "No sub 1.")
        self.assertEqual(grandchild.__doc__, "Sub 2 grandchild.")
        self.assertEqual(greatgrandchild.__doc__, "Sub 2 greatgrandchild.")

        class parent(metaclass=cfdm.core.meta.DocstringRewriteMeta):
            pass

        class child(parent):
            """{{class}}."""

        class grandchild(child):
            pass

        self.assertIsNone(parent.__doc__)
        self.assertEqual(child.__doc__, "child.")
        self.assertEqual(grandchild.__doc__, "grandchild.")

    def test_docstring(self):
        """Test that all double-brace markers have been substituted."""
        for klass in self.subclasses_of_Container:
            for x in (klass, klass()):
                for name in dir(x):
                    f = getattr(klass, name, None)

                    if f is None or not hasattr(f, "__doc__"):
                        continue

                    if name.startswith("__") and not inspect.isfunction(f):
                        continue

                    self.assertIsNotNone(
                        f.__doc__,
                        f"\n\nCLASS: {klass}\n"
                        f"METHOD NAME: {name}\n"
                        f"METHOD: {f}",
                    )

                    # if f.__doc__ is not None:
                    self.assertNotIn(
                        "{{",
                        f.__doc__,
                        f"\n\nCLASS: {klass}\n"
                        f"METHOD NAME: {name}\n"
                        f"METHOD: {f}",
                    )

    def test_docstring_package(self):
        """Test the docstring substitution of the package name."""
        string = f">>> f = {self.package}."
        for klass in self.subclasses_of_Container:
            for x in (klass, klass()):
                self.assertIn(string, x._has_component.__doc__, klass)

        string = f">>> f = {self.package}."
        for klass in self.subclasses_of_Properties:
            for x in (klass, klass()):
                self.assertIn(string, x.clear_properties.__doc__, klass)

    def test_docstring_class(self):
        """Test the docstring substitution of the class name."""
        for klass in self.subclasses_of_Properties:
            string = f">>> f = {self.package}.{klass.__name__}"
            for x in (klass, klass()):
                self.assertIn(
                    string,
                    x.clear_properties.__doc__,
                    f"\n\nCLASS: {klass}\n"
                    "METHOD NAME: clear_properties\n"
                    f"METHOD: {x.clear_properties}",
                )

        for klass in self.subclasses_of_Container:
            string = klass.__name__
            for x in (klass, klass()):
                self.assertIn(string, x.copy.__doc__, klass)

        for klass in self.subclasses_of_PropertiesDataBounds:
            string = f"{klass.__name__}"
            for x in (klass, klass()):
                self.assertIn(
                    string,
                    x.insert_dimension.__doc__,
                    f"\n\nCLASS: {klass}\n"
                    f"METHOD NAME: {klass.__name__}\n"
                    "METHOD: insert_dimension",
                )

    def test_docstring_repr(self):
        """Test docstring substitution of the object representation."""
        string = f"<{self.repr}Data"
        for klass in self.subclasses_of_PropertiesData:
            for x in (klass, klass()):
                self.assertIn(string, x.has_data.__doc__, klass)

    def test_docstring_default(self):
        """Test a given string gets substituted into a docstring."""
        string = "Return the value of the *default* parameter"
        for klass in self.subclasses_of_Properties:
            for x in (klass, klass()):
                self.assertIn(string, x.del_property.__doc__, klass)

    def test_docstring_staticmethod(self):
        """Test docstring substitution on a static method."""
        for klass in self.subclasses_of_PropertiesData:
            x = klass
            self.assertEqual(
                x._test_docstring_substitution_staticmethod(1, 2), (1, 2)
            )

    def test_docstring_classmethod(self):
        """Test docstring substitution on a class method."""
        for klass in self.subclasses_of_PropertiesData:
            for x in (klass, klass()):
                self.assertEqual(
                    x._test_docstring_substitution_classmethod(1, 2), (1, 2)
                )

    def test_docstring_docstring_substitutions(self):
        """Test the `_docstring substitution` internal method."""
        for klass in self.subclasses_of_Container:
            for x in (klass,):
                d = x._docstring_substitutions(klass)
                self.assertIsInstance(d, dict)
                self.assertIn("{{repr}}", d)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from ..h5netcdfarray import H5netcdfArray
from .mixin import FragmentFileArrayMixin


class FragmentH5netcdfArray(FragmentFileArrayMixin, H5netcdfArray):
    """A fragment of aggregated data in a file accessed with `h5netcdf`.

    .. versionadded:: (cfdm) 1.12.0.0

    """



import logging
import sys
from enum import Enum

import numpy as np
from dask import config
from dask.utils import parse_bytes

_CHUNKSIZE = "128 MiB"
config.set({"array.chunk-size": _CHUNKSIZE})

"""A dictionary of useful constants.

Whilst the dictionary may be modified directly, it is safer to
retrieve and set the values with the dedicated get-and-set functions.

:Keys:

    ATOL: `float`
      The value of absolute tolerance for testing numerically tolerant
      equality.

    RTOL: `float`
      The value of relative tolerance for testing numerically tolerant
      equality.

    LOG_LEVEL: `str`
      The minimal level of seriousness for which log messages are
      shown.  See `cfdm.log_level`.

    CHUNKSIZE: `int`
      The Dask chunk size (in bytes). See `cfdm.chunksize`.

"""
CONSTANTS = {
    "ATOL": sys.float_info.epsilon,
    "RTOL": sys.float_info.epsilon,
    "LOG_LEVEL": logging.getLevelName(logging.getLogger().level),
    "CHUNKSIZE": parse_bytes(_CHUNKSIZE),
}


# --------------------------------------------------------------------
# logging
# --------------------------------------------------------------------
class ValidLogLevels(Enum):
    """Enumerates all valid log levels for the logging in cfdm."""

    DISABLE = 0
    WARNING = 1
    INFO = 2
    DETAIL = 3
    DEBUG = -1


# --------------------------------------------------------------------
# masked
# --------------------------------------------------------------------
"""A constant that allows data values to be masked by direct
assignment. This is consistent with the behaviour of numpy masked
arrays.

For example, masking every element of a field constructs data array
could be done as follows:

>>> f[...] = cfdm.masked

"""
masked = np.ma.masked



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class CFDMImplementationTest(unittest.TestCase):
    """Unit test for the CFDMImplementation class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or calls (those
        # without a 'verbose' option to do the same) e.g. to debug them, wrap
        # them (for methods, start-to-end internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')
        self.i = cfdm.implementation()

    def test_CFDMImplementation__init__(self):
        """Test the constructor of CFDMImplementation."""
        cfdm.CFDMImplementation({"NewClass1": "qwerty", "NewClass2": None})

    def test_CFDMImplementation_classes(self):
        """Test the classes method of CFDMImplementation."""
        self.assertIsInstance(self.i.classes(), dict)

    def test_CFDMImplementation_set_class(self):
        """Test the `set_class` method of CFDMImplementation."""
        self.i.set_class("NewClass", "qwerty")

    def test_CFDMImplementation_get_class(self):
        """Test the `get_class` method of CFDMImplementation."""
        self.assertIs(self.i.get_class("Field"), cfdm.Field)

        with self.assertRaises(ValueError):
            self.i.get_class("qwerty")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import datetime
import faulthandler
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class SubsampledArrayTest(unittest.TestCase):
    """Unit test for the SubsampledArray class."""

    tie_point_indices = cfdm.TiePointIndex(data=[0, 4, 7, 8, 11])

    w = cfdm.InterpolationParameter(data=[5, 10, 5])

    coords = cfdm.SubsampledArray(
        interpolation_name="quadratic",
        compressed_array=cfdm.Data([15, 135, 225, 255, 345]),
        shape=(12,),
        tie_point_indices={0: tie_point_indices},
        parameters={"w": w},
        parameter_dimensions={"w": (0,)},
    )

    u_coords = np.array(
        [
            15.0,
            48.75,
            80.0,
            108.75,
            135.0,
            173.88888889,
            203.88888889,
            225.0,
            255.0,
            289.44444444,
            319.44444444,
            345.0,
        ]
    )

    bounds = cfdm.SubsampledArray(
        interpolation_name="quadratic",
        compressed_array=cfdm.Data([0, 150, 240, 240, 360]),
        shape=(12, 2),
        tie_point_indices={0: tie_point_indices},
        parameters={"w": w},
        parameter_dimensions={"w": (0,)},
    )
    u_bounds = np.array(
        [
            [0.0, 33.2],
            [33.2, 64.8],
            [64.8, 94.80000000000001],
            [94.80000000000001, 123.2],
            [123.2, 150.0],
            [150.0, 188.88888888888889],
            [188.88888888888889, 218.88888888888889],
            [218.88888888888889, 240.0],
            [240.0, 273.75],
            [273.75, 305.0],
            [305.0, 333.75],
            [333.75, 360.0],
        ]
    )

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_SubsampledArray__getitem__(self):
        """Test `SubsampledArray.__getitem__`"""
        coords = self.coords[...]
        bounds = self.bounds[...]
        self.assertEqual(coords.shape, self.u_coords.shape)
        self.assertEqual(bounds.shape, self.u_bounds.shape)
        self.assertTrue(np.allclose(coords, self.u_coords))
        self.assertTrue(np.allclose(bounds, self.u_bounds))

    def test_SubsampledArray_to_memory(self):
        """Test `SubsampledArray.to_memory."""
        self.assertIsInstance(self.coords.to_memory(), cfdm.SubsampledArray)

    def test_SubsampledArray_compressed_dimensions(self):
        """Test `SubsampledArray.compressed_dimensions."""
        self.assertEqual(self.coords.compressed_dimensions(), {0: (0,)})

        c = cfdm.SubsampledArray()
        with self.assertRaises(ValueError):
            c.compressed_dimensions()

    def test_SubsampledArray_get_filename(self):
        """Test SubsampledArray.get_filename."""
        x = self.coords
        self.assertIsNone(x.get_filename(default=None))

        with self.assertRaises(AttributeError):
            x.get_filename()


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



"""Define docstring substitutions.

Text to be replaced is specified as a key in the returned dictionary,
with the replacement text defined by the corresponding value.

Special docstring substitutions, as defined by a class's
`_docstring_special_substitutions` method, may be used in the
replacement text, and will be substituted as usual.

Replacement text may not contain other non-special substitutions.

Keys must be a `str` or `re.Pattern` object:

* If a key is a `str` then the corresponding value must be a string.

* If a key is a `re.Pattern` object then the corresponding value must
  be a string or a callable, as accepted by the `re.Pattern.sub`
  method.

.. versionaddedd:: (cfdm) 1.8.7.0

"""

_docstring_substitution_definitions = {
    # ----------------------------------------------------------------
    # General susbstitutions (not indent-dependent)
    # ----------------------------------------------------------------
    "{{repr}}": "",
    # ----------------------------------------------------------------
    # Method description susbstitutions (1 level of indentation)
    # ----------------------------------------------------------------
    # cell type
    "{{cell type}}": """The cell type indicates the common spatial dimensionality of
        the cells, one of:

        * ``'point'``: A point is zero-dimensional and has no boundary
                       vertices.
        * ``'edge'``: An edge is one-dimensional and corresponds to a
                      line connecting two boundary vertices.
        * ``'face'``: A face is two-dimensional and corresponds to a
                      surface enclosed by a set of edges.""",
    # cell connectivity type
    "{{cell connectivity type}}": """The connectivity type describes a characteristic of inter-cell
        connectivity defined by the domain topology construct. It may
        take any value, but the following values are standardised:

        * ``'node'``: Edge or face cells connected by one or more
                      shared nodes.
        * ``'edge'``: Face cells connected by one or more shared
                      edges.""",
    # ----------------------------------------------------------------
    # Method description susbstitutions (2 levels of indentation)
    # ----------------------------------------------------------------
    # cached: optional
    "{{cached: optional}}": """cached: optional
                If any value other than `None` then return *cached*
                without selecting any constructs.""",
    # todict: `bool`, optional
    "{{todict: `bool`, optional}}": """todict: `bool`, optional
                If True then return a dictionary of constructs keyed
                by their construct identifiers, instead of a
                `Constructs` object. This is a faster option.""",
    # ----------------------------------------------------------------
    # Method description susbstitutions (3 levels of indentation)
    # ----------------------------------------------------------------
    # axes int examples
    "{{axes int examples}}": """Each axis is identified by its integer position in the
                data. Negative integers counting from the last
                position are allowed.

                *Parameter example:*
                  ``axes=0``

                *Parameter example:*
                  ``axes=-1``

                *Parameter example:*
                  ``axes=[1, -2]``""",
    # default Exception
    "{{default Exception}}": """If set to an `Exception` instance then it will be
                raised instead.""",
    # inplace: `bool`, optional (default True)
    "{{inplace: `bool`, optional (default True)}}": """inplace: `bool`, optional:
                If True (the default) then do the operation in-place and
                return `None`. If False a new, modified `{{class}}`
                instance is returned.""",
    # init properties
    "{{init properties: `dict`, optional}}": """properties: `dict`, optional
                Set descriptive properties. The dictionary keys are
                property names, with corresponding values.

                Properties may also be set after initialisation with
                the `set_properties` and `set_property` methods.""",
    # init data
    "{{init data: data_like, optional}}": """data: data_like, optional
                Set the data.

                {{data_like}}

                The data also may be set after initialisation with the
                `set_data` method.""",
    # init bounds
    "{{init bounds: `Bounds`, optional}}": """bounds: `Bounds`, optional
                Set the bounds array.

                The bounds array may also be set after initialisation
                with the `set_bounds` method.""",
    # init geometry
    "{{init geometry: `str`, optional}}": """geometry: `str`, optional
                Set the geometry type.

                The geometry type may also be set after initialisation
                with the `set_geometry` method.

                *Parameter example:*
                  ``geometry='polygon'``""",
    # init interior_ring
    "{{init interior_ring: `InteriorRing`, optional}}": """interior_ring: `InteriorRing`, optional
                Set the interior ring variable.

                The interior ring variable may also be set after
                initialisation with the `set_interior_ring` method.""",
    # init copy
    "{{init copy: `bool`, optional}}": """copy: `bool`, optional
                If True (the default) then deep copy the input
                parameters prior to initialisation. By default the
                parameters are not deep copied.""",
    # init source
    "{{init source: optional}}": """source: optional
                Convert *source*, which can be any type of object, to
                a `{{class}}` instance.

                All other parameters, apart from *copy*, are ignored
                and their values are instead inferred from *source* by
                assuming that it has the `{{class}}` API. Any
                parameters that can not be retrieved from *source* in
                this way are assumed to have their default value.

                Note that if ``x`` is also a `{{class}}` instance then
                ``{{package}}.{{class}}(source=x)`` is equivalent to
                ``x.copy()``.""",
    # init cell
    "{{init cell: `str`, optional}}": """cell: `str`, optional
               The cell type that indicates the spatial dimensionality
               of the cells, one of ``'point'`` (a 0-d point with no
               boundary vertices), ``'edge'`` (a 1-d line connecting
               two boundary vertices), or ``'face'`` (a 2-d surface
               enclosed by a set of edges).""",
    # init connectivity
    "{{init connectivity: `str`, optional}}": """connectivity: `str`, optional
               The connectivity type describes a characteristic of
               inter-cell connectivity defined by the domain topology
               construct. It may take any value, but the following
               values are standardised: ``'node'``(edge or face cells
               connected by one or more shared nodes) and ``'edge'``
               (face cells connected by one or more shared edges).""",
    # data_like
    "{{data_like}}": """A data_like object is any object that can be converted
                to a `Data` object, i.e. `numpy` array_like objects,
                `Data` objects, and {{package}} instances that contain
                `Data` objects.""",
    # data: `bool`, optional
    "{{data: `bool`, optional}}": """data: `bool`, optional
                If True (the default) then copy data contained in the
                metadata construct(s), else the data is not copied.""",
    # (component-based) copy: `bool`, optional
    "{{copy: `bool`, optional}}": """copy: `bool`, optional
                If True (the default) then copy the component prior to
                insertion, else it is not copied.""",
    # data copy: `bool`, optional
    "{{data copy: `bool`, optional}}": """copy: `bool`, optional
                If True (the default) then copy the data prior to
                insertion, else the data is not copied.""",
}



import logging
import textwrap

from ..decorators import _display_or_return, _manage_log_level_via_verbosity
from . import Container

logger = logging.getLogger(__name__)


class Properties(Container):
    """Mixin class for descriptive properties.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return f"{self.identity('')}"

    def _dump_properties(self, _prefix="", _level=0, _omit_properties=None):
        """Dump the properties.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            _omit_properties: sequence of `str`, optional
                Omit the given CF properties from the description.

            _level: `int`, optional

        :Returns:

            `str`

        """
        indent0 = "    " * _level
        string = []

        properties = self.properties()

        if _omit_properties:
            for prop in _omit_properties:
                properties.pop(prop, None)

        for prop, value in sorted(properties.items()):
            name = f"{indent0}{_prefix}{prop} = "
            value = repr(value)
            subsequent_indent = " " * len(name)
            if value.startswith("'") or value.startswith('"'):
                subsequent_indent = f"{subsequent_indent} "

            string.append(
                textwrap.fill(
                    name + value, 79, subsequent_indent=subsequent_indent
                )
            )

        return "\n".join(string)

    def _identities_iter(self):
        """Return all possible identities.

        See `identities` for details and examples.

        :Returns:

            generator
                The identities.

        """
        standard_name = self.get_property("standard_name", None)
        if standard_name is not None:
            yield standard_name

        properties = self.properties()
        if properties:
            #            standard_name = properties.get("standard_name", None)
            #            if standard_name is not None:
            #                yield standard_name

            for prop in ("cf_role", "axis", "long_name"):
                value = properties.pop(prop, None)
                if value is not None:
                    yield f"{prop}={value}"

            for prop, value in sorted(properties.items()):
                yield f"{prop}={value}"

        ncvar = self.nc_get_variable(None)
        if ncvar is not None:
            yield f"ncvar%{ncvar}"

    def creation_commands(
        self, namespace=None, indent=0, string=True, name="c", header=True
    ):
        """Return the commands that would create the construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Domain.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.{{class}}(
        ...     properties={'units': 'Kelvin',
        ...                 'standard_name': 'air_temperature'}
        ... )
        >>> print(x.creation_commands(header=False))
        c = {{package}}.{{class}}()
        c.set_properties({'units': 'Kelvin', 'standard_name': 'air_temperature'})

        """
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        out = []

        if header:
            out.append("#")
            out.append("#")

            construct_type = getattr(self, "construct_type", None)
            if construct_type is not None:
                out[-1] += f" {construct_type}:"

            identity = self.identity()
            if identity:
                out[-1] += f" {identity}"

        out.append(f"{name} = {namespace}{self.__class__.__name__}()")

        properties = self.properties()
        if properties:
            out.append(f"{name}.set_properties({properties})")

        nc = self.nc_get_variable(None)
        if nc is not None:
            out.append(f"{name}.nc_set_variable('{nc}')")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    @_display_or_return
    def dump(
        self,
        display=True,
        _key=None,
        _omit_properties=(),
        _prefix="",
        _title=None,
        _create_title=True,
        _level=0,
    ):
        """A full description.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        indent0 = "    " * _level

        string = []

        # ------------------------------------------------------------
        # Title
        # ------------------------------------------------------------
        if _create_title:
            if _title is None:
                if _key:
                    default = f"key%{_key}"
                else:
                    default = ""

                string.append(
                    f"{indent0}{self.__class__.__name_}: "
                    f"{self.identity(default=default)}"
                )
            else:
                string.append(indent0 + _title)

        # ------------------------------------------------------------
        # Properties
        # ------------------------------------------------------------
        properties = self._dump_properties(
            _prefix=_prefix,
            _level=_level + 1,
            _omit_properties=_omit_properties,
        )
        if properties:
            string.append(properties)

        return "\n".join(string)

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_properties=None,
        ignore_type=False,
        ignore_compression=True,
    ):
        """Whether two instances are the same.

        Equality is strict by default. This means that:

        * the same descriptive properties must be present, with the
          same values and data types, and vector-valued properties
          must also have same the size and be element-wise equal (see
          the *ignore_properties* and *ignore_data_type* parameters).

        {{equals tolerance}}

        Any type of object may be tested but, in general, equality is
        only possible with another object of the same type, or a
        subclass of one. See the *ignore_type* parameter.

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{ignore_fill_value: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

            {{ignore_properties: (sequence of) `str`, optional}}

            {{ignore_data_type: `bool`, optional}}

            {{ignore_type: `bool`, optional}}

            ignore_compression:
                Ignored, since properties do not have data arrays.

        :Returns:

            `bool`
                Whether the two instances are equal.

        **Examples**

        >>> p.equals(p)
        True
        >>> p.equals(p.copy())
        True
        >>> p.equals('not a colection of properties')
        False

        >>> q = p.copy()
        >>> q.set_property('foo', 'bar')
        >>> p.equals(q)
        False
        >>> p.equals(q, verbose=3)
        {{class}}: Non-common property name: foo
        {{class}}: Different properties
        False

        """
        pp = super()._equals_preprocess(
            other, verbose=verbose, ignore_type=ignore_type
        )
        if pp is True or pp is False:
            return pp

        other = pp

        # ------------------------------------------------------------
        # Check the properties
        # ------------------------------------------------------------
        if ignore_fill_value:
            ignore_properties += ("_FillValue", "missing_value")

        self_properties = self.properties()
        other_properties = other.properties()

        if ignore_properties:
            if isinstance(ignore_properties, str):
                ignore_properties = (ignore_properties,)

            for prop in ignore_properties:
                self_properties.pop(prop, None)
                other_properties.pop(prop, None)

        if set(self_properties) != set(other_properties):
            for prop in set(self_properties).symmetric_difference(
                other_properties
            ):
                logger.info(
                    f"{self.__class__.__name__}: Missing property: {prop}"
                )

            return False

        for prop, x in self_properties.items():
            y = other_properties[prop]

            if not self._equals(
                x,
                y,
                rtol=rtol,
                atol=atol,
                ignore_fill_value=ignore_fill_value,
                ignore_data_type=True,
                verbose=verbose,
            ):
                logger.info(
                    f"{self.__class__.__name__}: Different {prop!r} "
                    f"property values: {x!r}, {y!r}"
                )

                return False

        return True

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The ``standard_name`` property.
        * The ``cf_role`` property, preceded by ``'cf_role='``.
        * The ``axis`` property, preceded by ``'axis='``.
        * The ``long_name`` property, preceded by ``'long_name='``.
        * The netCDF variable name, preceded by ``'ncvar%'``.
        * The value of the *default* parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_properties({'foo': 'bar',
        ...                   'long_name': 'Air Temperature',
        ...                   'standard_name': 'air_temperature'})
        >>> f.nc_set_variable('tas')
        >>> f.identity()
        'air_temperature'
        >>> f.del_property('standard_name')
        'air_temperature'
        >>> f.identity(default='no identity')
        'air_temperature'
        >>> f.identity()
        'long_name=Air Temperature'
        >>> f.del_property('long_name')
        >>> f.identity()
        'ncvar%tas'
        >>> f.nc_del_variable()
        'tas'
        >>> f.identity()
        'ncvar%tas'
        >>> f.identity()
        ''
        >>> f.identity(default='no identity')
        'no identity'

        """
        n = self.get_property("standard_name", None)
        if n is not None:
            return n

        for prop in ("cf_role", "axis", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The ``standard_name`` property.
        * All properties, preceded by the property name and an equals
          e.g. ``'long_name=Air temperature'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> f.properties()
        {'foo': 'bar',
         'long_name': 'Air Temperature',
         'standard_name': 'air_temperature'}
        >>> f.nc_get_variable()
        'tas'
        >>> f.identities()
        ['air_temperature',
         'long_name=Air Temperature',
         'foo=bar',
         'standard_name=air_temperature',
         'ncvar%tas']
        >>> for i in f.identities(generator=True):
        ...     print(i)
        ...
        air_temperature
        long_name=Air Temperature
        foo=bar
        standard_name=air_temperature
        ncvar%tas

        """
        g = self._iter(body=self._identities_iter(), **kwargs)
        if generator:
            return g

        return list(g)



import numpy as np

from ....core.utils import cached_property
from .subarray import Subarray


class SubsampledSubarray(Subarray):
    """A subarray of an array compressed by subsampling.

    A subarray describes a unique part of the uncompressed array, and
    corresponds to a single interpolation subarea, omitting any
    elements that are defined by previous (in index-space)
    interpolation subareas.

    See CF section 8.3 "Lossy Compression by Coordinate Subsampling"
    and appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __init__(
        self,
        data=None,
        indices=None,
        shape=None,
        compressed_dimensions=None,
        subarea_indices=None,
        first=None,
        parameters=None,
        dependent_tie_points=None,
        interpolation_description=None,
        source=None,
        copy=True,
        context_manager=None,
    ):
        """**Initialisation**

        :Parameters:

            data: array_like
                The full compressed tie points array spanning all
                subarrays, from which the elements for this subarray
                are defined by the *indices*.

            indices: `tuple`
                The indices of *data* that define this subarray.

            shape: `tuple` of `int`
                The shape of the uncompressed subarray.

            {{init compressed_dimensions: `dict`}}

                *Parameter example:*
                  ``{1: (1,)}``

                *Parameter example:*
                  ``{0: (0,), 2: (2,)}``

            first: `tuple` of `bool`
                For each dimension of the tie points array, True if
                the interpolation subarea is the first along that
                dimension of the continuous area, otherwise False.

            subarea_indices: `tuple` of `slice`
                For each dimension of the tie points array, the index
                that defines the location of the interpolation subarea
                in interpolation-subarea-space. An index corresponding
                to a non-interpolated dimension must be `slice(None)`.

            parameters: `dict`, optional
                If the interpolation method requires interpolation
                parameters then these are provided as array_like
                values, each keyed by its parameter term name. A
                parameter array must provide values for all
                interpolation subareas, from which the applicable
                elements are defined by the *subarea_indices* and
                *indices* indices.

                Each dimension of a parameter array maps to the
                dimension in the same position of the tie points
                array. The size of a parameter array dimension must be
                either i) the size as the corresponding tie points
                array dimension; ii) the size of the interpolation
                subarea dimension for the corresponding tie points
                array dimension; or iii) ``1``.

            dependent_tie_points: `dict`, optional
                If the interpolation method requires dependent tie
                point arrays then these are provided as array_like
                values, each keyed by its identity. A dependent tie
                points array must provide values for all interpolation
                subareas, from which the applicable elements are
                defined by the *indices* indices.

                A dependent tie points array must have the same shape
                as the tie points array, and each dimension maps to
                the dimension in the same position of the tie points
                array.

            interpolation_description: `str`, optional
                A complete description of the non-standardised
                interpolation method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

            context_manager: function, optional
                A context manager that provides a runtime context for
                the conversion of *data*, *dependent_tie_points*, and
                *parameters* to `numpy` arrays.

        """
        super().__init__(
            data=data,
            indices=indices,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            source=source,
            copy=copy,
            context_manager=context_manager,
        )

        if source is not None:
            try:
                subarea_indices = source._get_component(
                    "subarea_indices", None
                )
            except AttributeError:
                subarea_indices = None

            try:
                first = source._get_component("first", None)
            except AttributeError:
                first = None

            try:
                parameters = source._get_component("parameters", {})
            except AttributeError:
                parameters = {}

            try:
                dependent_tie_points = source._get_component(
                    "dependent_tie_points", {}
                )
            except AttributeError:
                dependent_tie_points = {}

            try:
                interpolation_description = source._get_component(
                    "interpolation_description", None
                )
            except AttributeError:
                interpolation_description = None

        if subarea_indices is not None:
            self._set_component("subarea_indices", subarea_indices, copy=copy)

        if first is not None:
            self._set_component("first", first, copy=False)

        if parameters is not None:
            self._set_component("parameters", parameters.copy(), copy=False)

        if dependent_tie_points is not None:
            self._set_component(
                "dependent_tie_points", dependent_tie_points.copy(), copy=False
            )

        if interpolation_description is not None:
            self._set_component(
                "interpolation_description",
                interpolation_description,
                copy=False,
            )

    def _broadcast_bounds(self, u):
        """Broadcast the raw uncompressed data to bounds locations.

        The raw uncompressed data is the basic interpolation of the
        bounds tie points, including the tie point locations.

        If *u* has not been calculated from bounds tie points then no
        broadcasting is done and the interpolated values are returned
        unchanged.

        See CF section 8.3.9 "Interpolation of Cell Boundaries".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_post_process`, `_trim`

        :Parameters:

            u: `numpy.ndarray`
                The raw uncompressed data is the basic interpolation
                of the tie points, including the tie point locations.

        :Returns:

            `numpy.ndarray`

        """
        if not self.bounds:
            return u

        if np.ma.isMA(u):
            bounds = np.ma.empty(self.shape, dtype=u.dtype)
        else:
            bounds = np.empty(self.shape, dtype=u.dtype)

        subsampled_dimensions = sorted(self.compressed_dimensions())
        n = len(subsampled_dimensions)

        indices = [slice(None)] * u.ndim

        if n == 1:
            (d1,) = subsampled_dimensions

            indices[d1] = slice(0, -1)
            bounds[..., 0] = u[tuple(indices)]

            indices[d1] = slice(1, None)
            bounds[..., 1] = u[tuple(indices)]
        elif n == 2:
            (d1, d2) = subsampled_dimensions

            indices[d1] = slice(0, -1)
            indices[d2] = slice(0, -1)
            bounds[..., 0] = u[tuple(indices)]

            indices[d2] = slice(1, None)
            bounds[..., 1] = u[tuple(indices)]

            indices[d1] = slice(1, None)
            bounds[..., 2] = u[tuple(indices)]

            indices[d2] = slice(1, None)
            indices[d2] = slice(0, -1)
            bounds[..., 3] = u[tuple(indices)]
        else:
            raise ValueError("Can only deal with 1 or 2 subsampled dimensions")

        return bounds

    def _codependent_tie_points(self, *identities):
        """Get all codependent tie points.

        Returns the tie points from `data` as well as those provided
        by `get_dependent_tie_points`.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            identities:
                The identities of all codependent tie points, of which
                all except one must be keys of the dictionary returned
                by `get_dependent_tie_points`. The identity that is
                not a key of that dictionary will correspond to the
                stored `tie_points`.

        :Returns:

            `list` of array_like
                The codependent tie points, in the order specified by
                the *identities* parameter.

        **Examples**

        >>> lat, lon = g._codependent_tie_points('latitude', 'longitude')

        """
        dependent_tie_points = self.dependent_tie_points

        if len(identities) != len(dependent_tie_points) + 1:
            raise ValueError(
                f"Must provide an identity for each of the "
                f"{len(dependent_tie_points) + 1} codependent tie point "
                f"arrays. Identities {identities} were provided, which "
                f"should include the identities {tuple(dependent_tie_points)}"
            )

        if not set(dependent_tie_points).issubset(identities):
            raise ValueError(
                "Each dependent tie point identity must be one of: "
                f"{', '.join(map(str, identities))}"
            )

        out = []
        for identity in identities:
            tie_points = dependent_tie_points.get(identity)
            if tie_points is None:
                out.append(self.data)
            else:
                out.append(tie_points)

        return out

    def _post_process(self, u):
        """Map the raw interpolated data to the uncompressed subarray.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            u: `numpy.ndarray`
               The uncompressed data for the whole interpolation
               subarea.

        :Returns:

            `numpy.ndarray`

        """
        u = self._broadcast_bounds(u)
        u = self._trim(u)
        return u

    def _s(self, d, s=None):
        """Return the interpolation coefficient ``s``.

        Returns the interpolation coefficient ``s`` for the specified
        subsampled dimension of the interpolation subarea. Note that
        the interpolation area may have fewer dimensions, and fewer
        elements, than the uncompresed subarray.

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{d: `int`}}

            {{s: array_like, optional}}

        :Returns:

            `numpy.ndarray`
                The interpolation coefficent ``s`` with values within
                the range [0, 1]. The array has extra size 1
                dimensions corresponding to all full tie point array
                dimensions other than *d*.

        **Examples**

        >>> x.shape
        (5, 5)
        >>> x.first
        (True, False)
        >>> x.bounds
        False
        >>> print(x._s(0))
        [[0.  ]
         [0.25]
         [0.5 ]
         [0.75]
         [1.  ]]
        >>> print(x._s(1))
        [[0.  0.2 0.4 0.6 0.8 1. ]]
        >>> print(x._s(1, s=0.5))
        [[0.5]]

        >>> x.shape
        (5, 5, 2)
        >>> x.first
        (True, False)
        >>> x.bounds
        True
        >>> print(x._s(0))
        [[0. ]
         [0.2]
         [0.4]
         [0.6]
         [0.8]
         [1. ]]
        >>> print(x._s(1))
        [[0.  0.2 0.4 0.6 0.8 1. ]]
        >>> print(x._s(1, s=0.5))
        [[0.5]]

        """
        ndim = self.data.ndim

        if s is not None:
            s = np.array(s, dtype=self.dtype)
            s.resize((1,) * ndim)
            return s

        size = self.shape[d]
        if self.bounds or not self.first[d]:
            size = size + 1

        s = np.linspace(0, 1, size, dtype=self.dtype)

        # Add extra size 1 dimensions so that s and 1-s are guaranteed
        # to be broadcastable to the tie points.
        if ndim > 1:
            new_shape = [1] * ndim
            new_shape[d] = size
            s.resize(new_shape)

        return s

    def _select_location(self, array, location=None):
        """Select interpolation parameter points interpolation subarea.

        Selects interpolation parameter points that correspond to this
        interpolation subarea.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_select_data`, `_select_parameter`

        :Parameters:

            array: `numpy.ndarray` or `None`
                All tie point or interpolation parameter values that
                apply to this interpolation subarea array, as returned
                by a call of `_select_data` or `_select_parameter`.

            location: `dict`, optional
                Identify the location within the interpolation
                subarea. Each key is an integer that specifies a
                subsampled dimension position in the array, with a
                value of either ``0`` or ``1`` indicating one of the
                two tie point positions along that dimension. By
                default, or if location is an empty dictionary, then
                all values for the interpolation subarea are returned.

        :Returns:

            `numpy.ndarray` or `None`
                The selected values, or `None` no array was provided.

        """
        if array is None:
            return

        if location:
            indices = [slice(None)] * array.ndim
            for subsampled_dimension, loc in location.items():
                indices[subsampled_dimension] = slice(loc, loc + 1)

            array = array[tuple(indices)]

        return array

    def _select_parameter(self, term):
        """Select interpolation parameter values.

        Selects interpolation parameter values for this interpolation
        subarea.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_select_data`, `_select_location`

        :Parameters:

            term: `str`
                The term name of an interpolation parameter. If *term*
                is not a key of the `parameters` dictionary, then
                `None` is returned.

        :Returns:

            `numpy.ndarray` or `None`
                The values of the interpolation parameter array that
                correspond to this interpolation subarea, or `None` if
                the interpolation parameter array has not been
                defined.

        """
        parameter = self.parameters.get(term)
        if parameter is None:
            return

        # Find the parameter array dimensions which are subsampled
        # dimensions
        subsampled_dimensions = [
            i
            for i, (m, n) in (enumerate(zip(parameter.shape, self.data.shape)))
            if m == n
        ]

        if subsampled_dimensions:
            indices = tuple(
                tp_index if dim in subsampled_dimensions else subarea_index
                for dim, (subarea_index, tp_index) in (
                    enumerate(zip(self.subarea_indices, self.indices))
                )
            )
        else:
            indices = self.subarea_indices

        return self._asanyarray(parameter, indices)

    def _trim(self, u):
        """Trim the raw uncompressed data.

        The raw uncompressed data is the basic interpolation of the
        tie points, including the tie point locations.

        For each subsampled dimension, removes the first point of the
        interpolation subarea when it is not the first (in
        index-space) of a continuous area. This is because this value
        in the uncompressed data has already been calculated from the
        previous (in index space) interpolation subarea.

        However, if *u* has been calculated from bounds tie points
        then no elements are removed. This is because all elements are
        used for broadcasting to each CF bounds location. See CF
        section 8.3.9 "Interpolation of Cell Boundaries".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_broadcast_bounds`, `_post_process`

        :Parameters:

            u: `numpy.ndarray`
                The raw uncompressed data is the basic interpolation
                of the tie points, including the tie point locations.

        :Returns:

            `numpy.ndarray`

        """
        if self.bounds:
            return u

        first = self.first

        take_slice = False
        indices = [slice(None)] * u.ndim
        for dim in self.compressed_dimensions():
            if first[dim]:
                continue

            take_slice = True
            indices[dim] = slice(1, None)

        if take_slice:
            u = u[tuple(indices)]

        return u

    @cached_property
    def bounds(self):
        """True if the tie points array represents bounds tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self.ndim > self.data.ndim

    @property
    def dependent_tie_points(self):
        """The dependent tie points dictionary.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("dependent_tie_points")

    @cached_property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return np.dtype(float)

    @property
    def first(self):
        """Relative interpolation subarea positions.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("first")

    @property
    def parameters(self):
        """The interpolation parameters dictionary.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("parameters")

    @property
    def subarea_indices(self):
        """Location of interpolation subarea.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("subarea_indices")

    def get_filename(self, normalise=True, default=AttributeError()):
        """Return the name of the file containing the data.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            {{normalise: `bool`, optional}}

                .. versionadded:: (cfdm) 1.12.0.0

            default: optional
                Return the value of the *default* parameter if there
                is no file name.

                {{default Exception}}

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `str`
                The file name.

        """
        filename = super().get_filename(normalise=normalise, default=None)
        if filename is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no unique file name"
            )

        filenames = [filename]
        for x in tuple(self.parameters.values()) + tuple(
            self.dependent_tie_points.values()
        ):
            try:
                filenames2 = x.get_filenames(normalise=normalise)
            except AttributeError:
                pass
            else:
                filenames.extend(filenames2)

        if len(set(filenames)) != 1:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no unique file name"
            )

        return filenames[0]

    def get_interpolation_description(self, default=ValueError()):
        """Get a non-standardised interpolation method description.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                interpolation description has not been set.

                {{default Exception}}

        """
        out = self._get_component("interpolation_description", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no "
                "'interpolation_description'",
            )

        return out



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class RaggedContiguousArrayTest(unittest.TestCase):
    """Unit test for the RaggedContiguousArray class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or calls (those
        # without a 'verbose' option to do the same) e.g. to debug them, wrap
        # them (for methods, start-to-end internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        compressed_data = cfdm.Data([280.0, 281.0, 279.0, 278.0, 279.5])
        count = cfdm.Count(data=[1, 3])
        self.r = cfdm.RaggedContiguousArray(
            compressed_data, shape=(2, 3), size=6, ndim=2, count_variable=count
        )

    def test_RaggedContiguousArray_to_memory(self):
        """Test `RaggedContiguousArray.to_memory`"""
        self.assertIsInstance(self.r.to_memory(), cfdm.RaggedContiguousArray)

    def test_RaggedContiguousArray_get_count(self):
        """Test `RaggedContiguousArray.get_count`"""
        r = self.r.copy()
        self.assertIsInstance(r.get_count(), cfdm.Count)
        r._del_component("count_variable")
        self.assertIsNone(r.get_count(None))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import atexit
import copy
import datetime
import faulthandler
import logging
import os
import platform
import sys
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_functions.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(temp_file,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class FunctionsTest(unittest.TestCase):
    """Test the functions of the non-core functions module."""

    @classmethod
    def setUpClass(cls):
        """Preparations called before the test class runs."""
        # Need to run this per-class, not per-method, to access the
        # original value of log_level to use to test the default (see
        # test_log_level)
        cls.original = cfdm.log_level("DISABLE")

        # When this module is run as part of full test-suite, the
        # value set in previous test is picked up above, not the
        # default. For now skip this test (effectively) for test suite
        # run. TODO: find way round this.
        if __name__ != "__main__":
            cls.original = "WARNING"

        # (no tearDownClass necessary)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warning, but
        # save original state for test on logging (see test_log_level)
        cfdm.log_level("DISABLE")
        # Note that test_log_level has been designed so the the above
        # call will not influence it, so it may be adjusted without
        # changing that test

        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        # Cover all below except lower-case, not supported in some
        # functions:
        valid_log_values = [-1, "INFO", 3, 2, 1, "DEBUG", "DETAIL", "WARNING"]
        # 'DISABLE' (0) is special case so exclude from levels list:
        self.valid_level_values = copy.copy(valid_log_values)
        # Cover string names, numeric code equivalents, & case
        # sensitivity:
        self.valid_log_values_ci = valid_log_values[:-2] + [
            "Detail",
            0,
            "DISABLE",
            "warning",
        ]

    def test_atol_rtol(self):
        """Test the atol and rtol functions."""
        org = cfdm.rtol()
        self.assertEqual(cfdm.rtol(1e-5), org)
        self.assertEqual(cfdm.rtol(), 1e-5)
        self.assertEqual(cfdm.rtol(org), 1e-5)
        self.assertEqual(cfdm.rtol(), org)

        org = cfdm.atol()
        self.assertEqual(cfdm.atol(1e-5), org)
        self.assertEqual(cfdm.atol(), 1e-5)
        self.assertEqual(cfdm.atol(org), 1e-5)
        self.assertEqual(cfdm.atol(), org)

        # Check aliases
        self.assertTrue(cfdm.atol() == cfdm.ATOL())
        self.assertTrue(cfdm.rtol() == cfdm.RTOL())

    def test_log_level(self):
        """Test the `log_level` function."""
        original = self.__class__.original  # original to module i.e. default

        self.assertEqual(original, "WARNING")  # test default
        cfdm.LOG_LEVEL(original)  # reset from setUp() value to avoid coupling

        # Now test getting and setting for all valid values in turn,
        # where use fact that setting returns old value hence set
        # value on next call:
        previous = cfdm.log_level()
        for value in self.valid_log_values_ci:
            self.assertEqual(cfdm.LOG_LEVEL(value), previous)
            previous = cfdm.LOG_LEVEL()  # update previous value

            # Some conversions to equivalent, standardised return value:
            if isinstance(value, int) and cfdm._is_valid_log_level_int(
                value
            ):  # str from LOG_LEVEL
                value = cfdm.constants.ValidLogLevels(value).name  # convert
            if isinstance(value, str):  # LOG_LEVEL returns all caps string
                value = value.upper()
            self.assertEqual(previous, value)

        with self.assertRaises(ValueError):
            cfdm.log_level(4)
            cfdm.LOG_LEVEL(4)  # check alias too
        with self.assertRaises(ValueError):
            cfdm.log_level("ERROR")  # notable as is valid Python logging level
            cfdm.LOG_LEVEL("ERROR")

    def test_reset_log_emergence_level(self):
        """Test the `_reset_log_emergence_level` internal function."""
        # 'DISABLE' is special case so test it afterwards (see below)
        for value in self.valid_level_values:
            cfdm.functions._reset_log_emergence_level(value)

            # getLevelName() converts to string. Otherwise gives
            # Python logging int equivalent, which is not the scale we use:
            if isinstance(value, int) and cfdm._is_valid_log_level_int(value):
                value = cfdm.constants.ValidLogLevels(value).name

            self.assertTrue(
                cfdm.logging.getLevelName(cfdm.logging.getLogger().level)
                == value
            )
        # Now test 'DISABLE' (0) special case; should not change level as such
        previous = logging.getLevelName(cfdm.logging.getLogger().level)
        cfdm.functions._reset_log_emergence_level("DISABLE")
        self.assertTrue(
            cfdm.logging.getLevelName(cfdm.logging.getLogger().level)
            == previous
        )
        # ... but test that it has disabled logging, as it is designed to do!
        self.assertFalse(cfdm.logging.getLogger().isEnabledFor(logging.DEBUG))
        self.assertFalse(
            cfdm.logging.getLogger().isEnabledFor(logging.WARNING)
        )

    def test_disable_logging(self):
        """Test the `_disable_logging` internal function."""
        # Re-set to avoid coupling; use set level to check it is
        # restored after
        cfdm.log_level("DETAIL")
        below_detail_values = [logging.DEBUG]
        at_or_above_detail_values = [
            cfdm.logging._nameToLevel["DETAIL"],
            logging.INFO,
            logging.WARNING,
        ]

        # Does it disable logging correctly?
        cfdm.functions._disable_logging()

        for value in below_detail_values + at_or_above_detail_values:
            self.assertFalse(cfdm.logging.getLogger().isEnabledFor(value))

        # And does it re-enable after having disabled logging if use
        # 'NOTSET'?
        cfdm.functions._disable_logging("NOTSET")  # should re-enable

        # Re-enabling should revert emergence in line with log
        # severity level:
        for value in at_or_above_detail_values:  # as long as level >= 'DETAIL'
            self.assertTrue(cfdm.logging.getLogger().isEnabledFor(value))
        # 'DEBUG' is effectively not "enabled" as is less severe than
        # 'DETAIL'
        self.assertFalse(cfdm.logging.getLogger().isEnabledFor(logging.DEBUG))

    def test_CF(self):
        """Test the CF function."""
        self.assertEqual(cfdm.CF(), cfdm.core.__cf_version__)

    def test_environment(self):
        """Test the `environment` function."""
        e = cfdm.environment(display=False)
        ep = cfdm.environment(display=False, paths=False)

        self.assertIsInstance(e, list)
        self.assertIsInstance(ep, list)

        components = ["Platform: ", "netCDF4: ", "numpy: ", "cftime: "]
        for component in components:
            self.assertTrue(any(s.startswith(component) for s in e))
            self.assertTrue(any(s.startswith(component) for s in ep))
        for component in [
            f"cfdm: {cfdm.__version__} {os.path.abspath(cfdm.__file__)}",
            f"Python: {platform.python_version()} {sys.executable}",
        ]:
            self.assertIn(component, e)
            self.assertNotIn(component, ep)  # paths shouldn't be present here
        for component in [
            f"cfdm: {cfdm.__version__}",
            f"Python: {platform.python_version()}",
        ]:
            self.assertIn(component, ep)

    def test_example_field(self):
        """Test the `example_field` function."""
        top = cfdm.examplefield._n_example_fields

        example_fields = cfdm.example_fields()
        self.assertEqual(len(example_fields), top)

        for n, f in enumerate(example_fields):
            _ = f.data.array

            self.assertIsInstance(f.dump(display=False), str)

            # TODOUGRID: When writing UGRID is enabled, we won't need
            #            the exclusion of fields 8, 9 and 10.
            if n not in (8, 9, 10):
                cfdm.write(f, temp_file)
                g = cfdm.read(temp_file, verbose=1)

                self.assertEqual(len(g), 1)
                self.assertTrue(f.equals(g[0], verbose=3))

        with self.assertRaises(Exception):
            cfdm.example_field(top + 1)

        with self.assertRaises(ValueError):
            cfdm.example_field(1, 2)

        with self.assertRaises(TypeError):
            cfdm.example_field(1, 2, 3)

        self.assertEqual(len(cfdm.example_fields(0)), 1)
        self.assertEqual(len(cfdm.example_fields(0, 2)), 2)
        self.assertEqual(len(cfdm.example_fields(0, 2, 0)), 3)

    def test_abspath(self):
        """Test cfdm.abspath."""
        cwd = os.getcwd()
        cwd_m1 = os.path.dirname(cwd)

        self.assertEqual(cfdm.abspath(""), cwd)
        self.assertEqual(cfdm.abspath("file.nc"), f"{cwd}/file.nc")
        self.assertEqual(cfdm.abspath("../file.nc"), f"{cwd_m1}/file.nc")
        self.assertEqual(cfdm.abspath("file:///file.nc"), "file:///file.nc")
        self.assertEqual(cfdm.abspath("file://file.nc"), f"file://{cwd}")
        self.assertEqual(cfdm.abspath("file:/file.nc"), "file:///file.nc")
        self.assertEqual(cfdm.abspath("http:///file.nc"), "http:///file.nc")
        self.assertEqual(cfdm.abspath("http://file.nc"), "http://")
        self.assertEqual(cfdm.abspath("http:/file.nc"), "http:///file.nc")

        self.assertEqual(
            cfdm.abspath("file.nc", uri=True), f"file://{cwd}/file.nc"
        )
        self.assertEqual(
            cfdm.abspath("../file.nc", uri=True), f"file://{cwd_m1}/file.nc"
        )
        self.assertEqual(
            cfdm.abspath("file:///file.nc", uri=True), "file:///file.nc"
        )
        self.assertEqual(
            cfdm.abspath("file://file.nc", uri=True), f"file://{cwd}"
        )
        self.assertEqual(
            cfdm.abspath("file:/file.nc", uri=True), "file:///file.nc"
        )
        self.assertEqual(
            cfdm.abspath("http:///file.nc", uri=True), "http:///file.nc"
        )
        self.assertEqual(cfdm.abspath("http://file.nc", uri=True), "http://")
        self.assertEqual(
            cfdm.abspath("http:/file.nc", uri=True), "http:///file.nc"
        )

        self.assertEqual(cfdm.abspath("file.nc", uri=False), f"{cwd}/file.nc")
        self.assertEqual(
            cfdm.abspath("../file.nc", uri=False), f"{cwd_m1}/file.nc"
        )
        self.assertEqual(
            cfdm.abspath("file:///file.nc", uri=False), "/file.nc"
        )
        self.assertEqual(cfdm.abspath("file://file.nc", uri=False), cwd)
        self.assertEqual(cfdm.abspath("file:/file.nc", uri=False), "/file.nc")
        with self.assertRaises(ValueError):
            cfdm.abspath("http:///file.nc", uri=False)

    def test_configuration(self):
        """Test the configuration function."""
        # Test getting of all config. and store original values to test on:
        org = cfdm.configuration()
        self.assertIsInstance(org, dict)
        self.assertEqual(len(org), 4)
        org_atol = org["atol"]
        self.assertIsInstance(org_atol, float)
        org_rtol = org["rtol"]
        self.assertIsInstance(org_rtol, float)
        org_ll = org["log_level"]  # will be 'DISABLE' as disable for test
        self.assertIsInstance(org_ll, str)
        org_chunksize = org["chunksize"]
        self.assertIsInstance(org_chunksize, int)

        # Store some sensible values to reset items to for testing,
        # ensure these are kept to be different to the defaults:
        atol_rtol_reset_value = 7e-10
        ll_reset_value = "DETAIL"

        # Test the setting of each lone item:
        cfdm.configuration(atol=atol_rtol_reset_value)
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], atol_rtol_reset_value)
        self.assertEqual(post_set["rtol"], org_rtol)
        self.assertEqual(post_set["log_level"], org_ll)
        cfdm.configuration(atol=org_atol)  # reset to org

        cfdm.configuration(rtol=atol_rtol_reset_value)
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], org_atol)
        self.assertEqual(post_set["rtol"], atol_rtol_reset_value)
        self.assertEqual(post_set["log_level"], org_ll)
        self.assertEqual(post_set["chunksize"], org_chunksize)
        # don't reset to org this time to test change persisting...

        # Note setting of previous items persist, e.g. atol above
        cfdm.configuration(log_level=ll_reset_value)
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], org_atol)
        self.assertEqual(
            post_set["rtol"], atol_rtol_reset_value
        )  # since changed it above
        self.assertEqual(post_set["log_level"], ll_reset_value)

        # Test the setting of more than one, but not all, items simultaneously:
        new_atol_rtol_reset_value = 5e-18
        new_ll_reset_value = "DEBUG"
        cfdm.configuration(
            rtol=new_atol_rtol_reset_value, log_level=new_ll_reset_value
        )
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], org_atol)
        self.assertEqual(post_set["rtol"], new_atol_rtol_reset_value)
        self.assertEqual(post_set["log_level"], new_ll_reset_value)

        # Test setting all possible items simultaneously (to originals):
        cfdm.configuration(
            atol=org_atol,  # same as current setting, testing on 'no change'
            rtol=org_rtol,
            log_level=org_ll,
        )
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], org_atol)
        self.assertEqual(post_set["rtol"], org_rtol)
        self.assertEqual(post_set["log_level"], org_ll)

        # Test edge cases & invalid inputs...
        # ... 1. User might set '0' or 'True' in some cases, which is
        # somewhat a risk area for error as 0 is Falsy & True a Bool:
        cfdm.configuration(rtol=0, atol=0.0, log_level=0)
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], 0.0)
        self.assertEqual(post_set["rtol"], 0.0)
        self.assertEqual(post_set["log_level"], "DISABLE")  # DISABLE == 0
        cfdm.configuration(log_level=True)  # deprecated but valid value
        # Deprecated True is converted to value 'WARNING' by log_level
        self.assertEqual(cfdm.configuration()["log_level"], "WARNING")

        # 2. None as an input kwarg rather than as a default:
        cfdm.configuration(atol=None, log_level=None, rtol=org_rtol)
        post_set = cfdm.configuration()
        self.assertEqual(post_set["atol"], 0.0)  # 0.0 as set above
        self.assertEqual(post_set["rtol"], org_rtol)
        self.assertEqual(post_set["log_level"], "WARNING")  # as set above

        # 3. Gracefully error with useful error messages with invalid inputs:
        with self.assertRaises(ValueError):
            cfdm.configuration(rtol="bad")
        with self.assertRaises(ValueError):
            cfdm.configuration(log_level=7)

        # 4. Check invalid kwarg given logic processes **kwargs:
        with self.assertRaises(TypeError):
            cfdm.configuration(bad_kwarg=1e-15)

        old = cfdm.configuration()
        try:
            cfdm.configuration(atol=888, rtol=999, log_level="BAD")
        except ValueError:
            self.assertEqual(cfdm.configuration(), old)
        else:
            raise RuntimeError(
                "A ValueError should have been raised, but wasn't"
            )

        # Reset configuration
        cfdm.configuration(**org)
        self.assertEqual(cfdm.configuration(), org)

    def test_unique_constructs(self):
        """Test the `unique_constructs` function."""
        f = cfdm.example_field(0)
        g = cfdm.example_field(1)

        self.assertFalse(cfdm.unique_constructs([]))

        self.assertEqual(len(cfdm.unique_constructs([f])), 1)
        self.assertEqual(len(cfdm.unique_constructs([f, f])), 1)
        self.assertEqual(len(cfdm.unique_constructs([f, f.copy()])), 1)
        self.assertEqual(len(cfdm.unique_constructs([f, f.copy(), g])), 2)

        fields = [f, f, g]
        domains = [x.domain for x in (f, f, g)]

        u = cfdm.unique_constructs(domains)
        self.assertEqual(len(u), 2)
        self.assertIsNot(u[0], domains[0])

        u = cfdm.unique_constructs(domains, copy=False)
        self.assertIs(u[0], domains[0])

        u = cfdm.unique_constructs(domains, ignore_properties="z", copy=False)
        self.assertIsNot(u[0], domains[0])

        self.assertEqual(len(cfdm.unique_constructs(domains + fields)), 4)
        self.assertEqual(
            len(cfdm.unique_constructs(domains + fields + [f.domain])), 4
        )

        # Test generator
        domains = (x.domain for x in ())
        u = cfdm.unique_constructs(domains)
        self.assertIsInstance(u, type(domains))
        self.assertEqual(len(tuple(u)), 0)

        domains = (x.domain for x in (f,))
        u = cfdm.unique_constructs(domains)
        self.assertIsInstance(u, type(domains))
        self.assertEqual(len(tuple(u)), 1)

        domains = (x.domain for x in (f, f, g))
        u = cfdm.unique_constructs(domains)
        self.assertIsInstance(u, type(domains))
        self.assertEqual(len(tuple(u)), 2)

        self.assertIsInstance(cfdm.unique_constructs(fields), type(fields))
        self.assertIsInstance(cfdm.unique_constructs(tuple(fields)), tuple)

        f.set_properties({"project": "model", "foo": "bar"})
        f.set_property("project", "model")
        f2 = f.copy()
        f2.set_property("project", "obs")
        g.set_properties({"project": "model", "foo": "bar"})
        fields = [f, f2, g]

        self.assertEqual(len(cfdm.unique_constructs(fields)), 3)
        u = cfdm.unique_constructs(
            fields, ignore_properties=["project", "foo"]
        )
        self.assertEqual(len(u), 2)
        self.assertIsNone(u[0].get_property("project", None))
        self.assertEqual(u[0].get_property("foo"), "bar")
        self.assertEqual(u[1].get_property("project"), "model")
        self.assertEqual(u[1].get_property("foo"), "bar")

    def test_context_managers(self):
        """Test the context manager support of the functions."""
        # rtol and atol
        for func in (cfdm.atol, cfdm.rtol):
            old = func()
            new = old * 2
            with func(new):
                self.assertEqual(func(), new)
                self.assertEqual(func(new * 2), new)
                self.assertEqual(func(), new * 2)

            self.assertEqual(func(), old)

        # log_level
        func = cfdm.log_level

        org = func("DETAIL")
        old = func()
        new = "DEBUG"
        with func(new):
            self.assertEqual(func(), new)

        self.assertEqual(func(), old)
        func(org)

        del org._func
        with self.assertRaises(AttributeError):
            with org:
                pass

        # Full configuration
        func = cfdm.configuration

        org = func(rtol=10, atol=20, log_level="DETAIL")
        old = func()
        new = dict(old)
        new["rtol"] = 10 * 2
        new["atol"] = 20 * 2
        new["log_level"] = "DEBUG"

        with func(**new):
            self.assertEqual(func(), new)

        self.assertEqual(func(), old)
        func(**org)

        org = func(rtol=cfdm.Constant(10), atol=20, log_level="DETAIL")
        old = func()
        new["rtol"] = cfdm.Constant(10 * 2)
        new["atol"] = 20 * 2
        new["log_level"] = "DEBUG"

        with func(**new):
            self.assertEqual(func(), new)

        self.assertEqual(func(), old)
        func(**org)

    def test_Constant(self):
        """Test the Constant class."""
        c = cfdm.Constant(20)
        d = cfdm.Constant(10)
        e = cfdm.Constant(999)

        self.assertIsInstance(hash(c), int)
        self.assertIsInstance(repr(c), str)

        self.assertEqual(float(c), 20.0)
        self.assertEqual(int(c), 20)
        self.assertEqual(str(c), "20")

        # Binary operations
        self.assertEqual(c, 20)
        self.assertEqual(20, c)
        self.assertEqual(c, c)
        self.assertEqual(c, copy.deepcopy(c))
        self.assertEqual(c, np.array(20))

        self.assertNotEqual(c, 999)
        self.assertNotEqual(999, c)
        self.assertNotEqual(c, d)
        self.assertNotEqual(c, np.array(999))
        self.assertNotEqual(np.array(999), c)

        self.assertLess(c, 999)
        self.assertLessEqual(c, 999)
        self.assertGreater(c, 10)
        self.assertGreaterEqual(c, 10)

        self.assertGreater(999, c)
        self.assertGreaterEqual(999, c)
        self.assertLess(10, c)
        self.assertLessEqual(10, c)

        self.assertLess(c, e)
        self.assertLessEqual(c, e)
        self.assertGreater(c, d)
        self.assertGreaterEqual(c, d)

        self.assertEqual(c + 10, 30)
        self.assertEqual(c - 10, 10)
        self.assertEqual(c / 10, 2)
        self.assertEqual(c * 10, 200)
        self.assertEqual(c // 10, 2)

        self.assertEqual(c + d, 30)
        self.assertEqual(c - d, 10)
        self.assertEqual(c / d, 2)
        self.assertEqual(c * d, 200)
        self.assertEqual(c // d, 2)

        self.assertEqual(20 + d, 30)
        self.assertEqual(20 - d, 10)
        self.assertEqual(20 / d, 2)
        self.assertEqual(20 * d, 200)
        self.assertEqual(20 // d, 2)

        c = cfdm.Constant(20)
        c -= 10
        self.assertEqual(c, 10)
        c += 10
        self.assertEqual(c, 20)
        c *= 10
        self.assertEqual(c, 200)
        c /= 10
        self.assertEqual(c, 20)
        c //= 10
        self.assertEqual(c, 2)

        # Unary operations
        c = cfdm.Constant(-20)
        self.assertEqual(-c, 20)
        self.assertEqual(abs(c), 20)
        self.assertEqual(+c, -20)

        # Copy
        c = cfdm.atol().copy()
        del c._func
        self.assertEqual(c, c.copy())

        # Bool
        self.assertTrue(cfdm.Constant(1))
        self.assertTrue(cfdm.Constant(True))
        self.assertFalse(cfdm.Constant(0))
        self.assertFalse(cfdm.Constant(False))

    def test_Configuration(self):
        """Test the Configuration class."""
        c = cfdm.configuration()

        self.assertIsInstance(repr(c), str)
        self.assertEqual(str(c), str(dict(**c)))

    def test_dirname(self):
        """Test cfdm.dirname."""
        cwd = os.getcwd()
        cwd_m1 = os.path.dirname(cwd)

        self.assertEqual(cfdm.dirname("file.nc"), "")
        self.assertEqual(cfdm.dirname("file.nc", normalise=True), cwd)
        self.assertEqual(
            cfdm.dirname("file.nc", normalise=True, uri=True), f"file://{cwd}"
        )
        self.assertEqual(
            cfdm.dirname("file.nc", normalise=True, uri=False), cwd
        )
        self.assertEqual(
            cfdm.dirname("file.nc", normalise=True, sep=True), f"{cwd}/"
        )

        self.assertEqual(cfdm.dirname("model/file.nc"), "model")
        self.assertEqual(
            cfdm.dirname("model/file.nc", normalise=True), f"{cwd}/model"
        )
        self.assertEqual(
            cfdm.dirname("model/file.nc", normalise=True, uri=True),
            f"file://{cwd}/model",
        )
        self.assertEqual(
            cfdm.dirname("model/file.nc", normalise=True, uri=False),
            f"{cwd}/model",
        )

        self.assertEqual(cfdm.dirname("../file.nc"), "..")
        self.assertEqual(cfdm.dirname("../file.nc", normalise=True), cwd_m1)
        self.assertEqual(
            cfdm.dirname("../file.nc", normalise=True, uri=True),
            f"file://{cwd_m1}",
        )
        self.assertEqual(
            cfdm.dirname("../file.nc", normalise=True, uri=False), cwd_m1
        )

        self.assertEqual(cfdm.dirname("/model/file.nc"), "/model")
        self.assertEqual(
            cfdm.dirname("/model/file.nc", normalise=True), "/model"
        )
        self.assertEqual(
            cfdm.dirname("/model/file.nc", normalise=True, uri=True),
            "file:///model",
        )
        self.assertEqual(
            cfdm.dirname("/model/file.nc", normalise=True, uri=False), "/model"
        )

        self.assertEqual(cfdm.dirname(""), "")
        self.assertEqual(cfdm.dirname("", normalise=True), cwd)
        self.assertEqual(
            cfdm.dirname("", normalise=True, uri=True), f"file://{cwd}"
        )
        self.assertEqual(cfdm.dirname("", normalise=True, uri=False), cwd)

        self.assertEqual(
            cfdm.dirname("https:///data/archive/file.nc"),
            "https:///data/archive",
        )
        self.assertEqual(
            cfdm.dirname("https:///data/archive/file.nc", normalise=True),
            "https:///data/archive",
        )
        self.assertEqual(
            cfdm.dirname(
                "https:///data/archive/file.nc", normalise=True, uri=True
            ),
            "https:///data/archive",
        )
        with self.assertRaises(ValueError):
            cfdm.dirname(
                "https:///data/archive/file.nc", normalise=True, uri=False
            )

        self.assertEqual(
            cfdm.dirname("file:///data/archive/file.nc"),
            "file:///data/archive",
        )
        self.assertEqual(
            cfdm.dirname("file:///data/archive/file.nc", normalise=True),
            "file:///data/archive",
        )
        self.assertEqual(
            cfdm.dirname(
                "file:///data/archive/file.nc", normalise=True, uri=True
            ),
            "file:///data/archive",
        )
        self.assertEqual(
            cfdm.dirname(
                "file:///data/archive/file.nc", normalise=True, uri=False
            ),
            "/data/archive",
        )

        self.assertEqual(
            cfdm.dirname("file:///data/archive/../file.nc"),
            "file:///data/archive/..",
        )
        self.assertEqual(
            cfdm.dirname("file:///data/archive/../file.nc", normalise=True),
            "file:///data",
        )
        self.assertEqual(
            cfdm.dirname(
                "file:///data/archive/../file.nc", normalise=True, uri=True
            ),
            "file:///data",
        )
        self.assertEqual(
            cfdm.dirname(
                "file:///data/archive/../file.nc", normalise=True, uri=False
            ),
            "/data",
        )


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import numpy as np

from .quadraticgeographicinterpolation import QuadraticGeographicInterpolation


class BiQuadraticLatitudeLongitudeInterpolation(
    QuadraticGeographicInterpolation
):
    """Mixin class for bi-quadratic latitude longitude interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _bi_quadratic_latitude_longitude_interpolation(
        self,
        lat_a,
        lon_a,
        lat_b,
        lon_b,
        lat_c,
        lon_c,
        lat_d,
        lon_d,
        ce1,
        ca1,
        ce2,
        ca2,
        ce3,
        ca3,
        location_use_3d_cartesian,
        d2,
        d1,
    ):
        """Bi-quadratic interpolation of geographic coordinates.

        A two-dimensional quadratic method for interpolation of the
        geographic coordinates latitude and longitude, typically used
        for remote sensing products with geographic coordinates on the
        reference ellipsoid.

        See CF appendix J "Coordinate Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            lat_a, lon_a: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location A, in the sense of CF appendix J Figure J.2.

            lat_b, lon_b: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location B, in the sense of CF appendix J Figure J.2.

            lat_c, lon_c: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location C, in the sense of CF appendix J Figure J.2.

            lat_d, lon_d: `numpy.ndarray`
                The latitude and longitude of the tie point at
                location D, in the sense of CF appendix J Figure J.2.

            ce1, ca1, ce2, ca2, ce3, ca3: `numpy.ndarray` or `None`
                The interpolation parameters ``ce1``, ``ca1``,
                ``ce2``, ``ca2``, ``ce3``, and ``ca3``, with the same
                number of dimensions in the same relative order as the
                full tie points array. If any are `None` then those
                parameter are assumed to be zero.

            {{location_use_3d_cartesian: `numpy.ndarray` or `None`}}

            {{d2: `int`}}

            {{d1: `int`}}

        :Returns:

            `numpy.ndarray`

        """
        if location_use_3d_cartesian is None:
            raise ValueError(
                "Can't uncompress tie points by "
                "bi_quadratic_latitude_longitude interpolation "
                "when there is no interpolation_subarea_flags "
                "interpolation parameter that includes "
                "'location_use_3d_cartesian' in its flag_meanings "
                "attribute"
            )

        any_cartesian = bool(location_use_3d_cartesian.any())
        all_cartesian = bool(location_use_3d_cartesian.all())

        # Find out if these tie points are latitudes or longitudes. If
        # they are latitudes then the dependent tie points will be
        # longitudes.
        latitude = "longitude" in self.dependent_tie_points

        if latitude:
            fv2ll = self._fv2lat
        else:
            fv2ll = self._fv2lon

        va = self._fll2v(lat_a, lon_a)
        vb = self._fll2v(lat_b, lon_b)
        vc = self._fll2v(lat_c, lon_c)
        vd = self._fll2v(lat_d, lon_d)

        # ce2, ca2: Span interpolation subarea dimension 2 and
        #           subsampled dimension 1
        cea2 = (
            self._select_location(ce2, {d1: 0}),
            self._select_location(ca2, {d1: 0}),
        )
        cv_ac = self._fcea2cv(va, vc, *cea2)

        cea2 = (
            self._select_location(ce2, {d1: 1}),
            self._select_location(ca2, {d1: 1}),
        )
        cv_bd = self._fcea2cv(vb, vd, *cea2)
        del cea2

        # ce1, ca1: Span subsampled dimension 2 and interpolation
        #           subarea dimension 1
        cea1 = (
            self._select_location(ce1, {d2: 0}),
            self._select_location(ca1, {d2: 0}),
        )
        vab = self._fqv(va, vb, self._fcea2cv(va, vb, *cea1), d1, s=0.5)

        cea1 = (
            self._select_location(ce1, {d2: 1}),
            self._select_location(ca1, {d2: 1}),
        )
        vcd = self._fqv(vc, vd, self._fcea2cv(vc, vd, *cea1), d1, s=0.5)
        del cea1

        # ce3, ca3: Span interpolation subarea dimension 2 and
        #           interpolation subarea dimension 1
        cea3 = (self._select_location(ce3, {}), self._select_location(ca3, {}))
        cv_z = self._fcea2cv(vab, vcd, *cea3)
        del cea3

        if any_cartesian:
            # Interpolation in three-dimensional cartesian coordinates
            vac = self._fqv(va, vc, cv_ac, d2)
            if all_cartesian:
                del va, vc, cv_ac

            vbd = self._fqv(vb, vd, cv_bd, d2)
            if all_cartesian:
                del vb, vd, cv_bd

            vz = self._fqv(vab, vcd, cv_z, d2)
            if all_cartesian:
                del vab, vcd, cv_z

            cv_zz = self._fcv(vac, vbd, vz, d1, s_i=0.5)
            del vz

            vp = self._fqv(vac, vbd, cv_zz, d1)
            del vac, vbd, cv_zz

            u_c = fv2ll(vp)
            del vp

            if all_cartesian:
                llp = u_c

        if not all_cartesian:
            # Interpolation in latitude-longitude coordinates
            if latitude:
                lla, llb, llc, lld = lat_a, lat_b, lat_c, lat_d
            else:
                lla, llb, llc, lld = lon_a, lon_b, lon_c, lon_d

            llc_ac = self._fw(
                lla,
                llc,
                fv2ll(self._fqv(va, vc, cv_ac, d2, s=0.5)),
                d2,
                s_i=0.5,
            )
            del va, vc, cv_ac

            llac = self._fq(lla, llc, llc_ac, d2)
            del lla, llc, llc_ac
            if latitude:
                del lat_a, lat_c
            else:
                del lon_a, lon_c

            llc_bd = self._fw(
                llb,
                lld,
                fv2ll(self._fqv(vb, vd, cv_bd, d2, s=0.5)),
                d2,
                s_i=0.5,
            )
            del vb, vd, cv_bd

            llbd = self._fq(llb, lld, llc_bd, d2)
            del llb, lld, llc_bd
            if latitude:
                del lat_b, lat_d
            else:
                del lon_b, lon_d

            llab = fv2ll(vab)
            llcd = fv2ll(vcd)
            llc_z = self._fw(
                llab,
                llcd,
                fv2ll(self._fqv(vab, vcd, cv_z, d2, s=0.5)),
                d2,
                s_i=0.5,
            )
            del vab, vcd, cv_z

            llz = self._fq(llab, llcd, llc_z, d2)
            del llc_z

            cl_zz = self._fw(llac, llbd, llz, d1, s_i=0.5)
            del llz

            u_l = self._fq(llac, llbd, cl_zz, d1)
            del llac, llbd, cl_zz

            if not any_cartesian:
                llp = u_l

        if any_cartesian and not all_cartesian:
            # Combine the results of cartesian and latitude-longitude
            # interpolations
            llp = np.where(location_use_3d_cartesian, u_c, u_l)

        return llp



class DeprecationError(Exception):
    """Deprecation error."""

    pass


class NetCDFArray:
    """A netCDF array accessed with `netCDF4`.

    Deprecated at version 1.11.2.0 and is no longer available. Use
    `cfdm.NetCDF4Array` instead.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        shape=None,
        mask=True,
        unpack=True,
        units=False,
        calendar=False,
        attributes=None,
        storage_options=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: (sequence of) `str`, optional
                The name of the netCDF file(s) containing the array.

            address: (sequence of) `str` or `int`, optional
                The identity of the netCDF variable in each file
                defined by *filename*. Either a netCDF variable name
                or an integer netCDF variable ID.

                .. versionadded:: (cfdm) 1.10.1.0

            dtype: `numpy.dtype`
                The data type of the array in the netCDF file. May be
                `None` if the numpy data-type is not known (which can be
                the case for netCDF string types, for example).

            shape: `tuple`
                The array dimension sizes in the netCDF file.

            size: `int`
                Number of elements in the array in the netCDF file.

            ndim: `int`
                The number of array dimensions in the netCDF file.

            {{init mask: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.2

            units: `str` or `None`, optional
                The units of the netCDF variable. Set to `None` to
                indicate that there are no units. If unset then the
                units will be set during the first `__getitem__` call.

                .. versionadded:: (cfdm) 1.10.0.1

            calendar: `str` or `None`, optional
                The calendar of the netCDF variable. By default, or if
                set to `None`, then the CF default calendar is
                assumed, if applicable. If unset then the calendar
                will be set during the first `__getitem__` call.

                .. versionadded:: (cfdm) 1.10.0.1

            {{init source: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{init copy: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            missing_values: Deprecated at version 1.11.2.0
                The missing value indicators defined by the netCDF
                variable attributes. They may now be recorded via the
                *attributes* parameter

            ncvar:  Deprecated at version 1.10.1.0
                Use the *address* parameter instead.

            varid:  Deprecated at version 1.10.1.0
                Use the *address* parameter instead.

            group: Deprecated at version 1.10.1.0
                Use the *address* parameter instead.

        """
        raise DeprecationError(
            f"{self.__class__.__name__} was deprecated at version 1.11.2.0 "
            "and is no longer available. Use cfdm.NetCDF4Array instead."
        )



import logging

from . import CoordinateConversion, Datum, core, mixin
from .data import Data
from .decorators import _display_or_return, _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class CoordinateReference(
    mixin.NetCDFVariable,
    mixin.Container,
    mixin.Files,
    core.CoordinateReference,
):
    """A coordinate reference construct of the CF data model.

    A coordinate reference construct relates the coordinate values of
    the coordinate system to locations in a planetary reference frame.

    The domain of a field construct may contain various coordinate
    systems, each of which is constructed from a subset of the
    dimension and auxiliary coordinate constructs. For example, the
    domain of a four-dimensional field construct may contain
    horizontal (y-x), vertical (z), and temporal (t) coordinate
    systems. There may be more than one of each of these, if there is
    more than one coordinate construct applying to a particular
    spatiotemporal dimension (for example, there could be both
    latitude-longitude and y-x projection coordinate systems). In
    general, a coordinate system may be constructed implicitly from
    any subset of the coordinate constructs, yet a coordinate
    construct does not need to be explicitly or exclusively associated
    with any coordinate system.

    A coordinate system of the field construct can be explicitly
    defined by a coordinate reference construct which relates the
    coordinate values of the coordinate system to locations in a
    planetary reference frame and consists of the following:

    * References to the dimension coordinate and auxiliary coordinate
      constructs that define the coordinate system to which the
      coordinate reference construct applies. Note that the coordinate
      values are not relevant to the coordinate reference construct,
      only their properties.

    ..

    * A definition of a datum specifying the zeroes of the dimension
      and auxiliary coordinate constructs which define the coordinate
      system. The datum may be implied by the metadata of the
      referenced dimension and auxiliary coordinate constructs, or
      explicitly provided.

    ..

    * A coordinate conversion, which defines a formula for converting
      coordinate values taken from the dimension or auxiliary
      coordinate constructs to a different coordinate system. A
      coordinate reference construct relates the coordinate values of
      the field to locations in a planetary reference frame.

    **NetCDF interface**

    {{netCDF variable}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._CoordinateConversion = CoordinateConversion
        instance._Data = Data
        instance._Datum = Datum
        return instance

    def __init__(
        self,
        coordinates=None,
        datum=None,
        coordinate_conversion=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            coordinates: sequence of `str`, optional
                Identify the related dimension and auxiliary
                coordinate constructs by their construct
                identifiers.

                The coordinates may also be set after initialisation
                with the `set_coordinates` and `set_coordinate`
                methods.

                *Parameter example:*
                  ``coordinates=['dimensioncoordinate2']``

                *Parameter example:*
                  ``coordinates=('dimensioncoordinate0', 'dimensioncoordinate1')``

            datum: `Datum`, optional
                Set the datum component of the coordinate reference
                construct.

                The datum may also be set after initialisation with
                the `set_datum` method.

            coordinate_conversion: `CoordinateConversion`, optional
                Set the coordinate conversion component of the
                coordinate reference construct.

                The coordinate conversion may also be set after
                initialisation with the `set_coordinate_conversion`
                method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            coordinates=coordinates,
            datum=datum,
            coordinate_conversion=coordinate_conversion,
            source=source,
            copy=copy,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return self.identity(default=self.nc_get_variable(""))

    def _identities_iter(self):
        """Return all possible identities.

        See `identities` for details and examples.

        :Returns:

            generator
                The identities.

        """
        cc = self.coordinate_conversion
        for prop in ("standard_name", "grid_mapping_name"):
            n = cc.get_parameter(prop, None)
            if n is not None:
                yield f"{prop}:{n}"

        ncvar = self.nc_get_variable(None)
        if ncvar is not None:
            yield f"ncvar%{ncvar}"

    def creation_commands(
        self, namespace=None, indent=0, string=True, name="c", header=True
    ):
        """Returns the commands to create the coordinate reference.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{header: `bool`, optional}}

            {{name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.{{class}}(
        ...     coordinates=['dimensioncoordinate0']
        ... )
        >>> x.datum.set_parameter('earth_radius', 6371007)
        >>> x.coordinate_conversion.set_parameters(
        ...     {'standard_name': 'atmosphere_hybrid_height_coordinate',
        ...      'computed_standard_name': 'altitude'}
        ... )
        >>> x.coordinate_conversion.set_domain_ancillaries(
        ...     {'a': 'domainancillary0',
        ...      'b': 'domainancillary1',
        ...      'orog': 'domainancillary2'}
        ... )
        >>> print(x.creation_commands(header=False))
        c = {{package}}.{{class}}()
        c.set_coordinates({'dimensioncoordinate0'})
        c.datum.set_parameter('earth_radius', 6371007)
        c.coordinate_conversion.set_parameter('standard_name', 'atmosphere_hybrid_height_coordinate')
        c.coordinate_conversion.set_parameter('computed_standard_name', 'altitude')
        c.coordinate_conversion.set_domain_ancillaries({'a': 'domainancillary0', 'b': 'domainancillary1', 'orog': 'domainancillary2'})

        """
        namespace0 = namespace
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        out = []

        if header:
            out.append("#")
            out.append(f"# {self.construct_type}:")
            identity = self.identity()
            if identity:
                out[-1] += f" {identity}"

        out.append(f"{name} = {namespace}{self.__class__.__name__}()")

        nc = self.nc_get_variable(None)
        if nc is not None:
            out.append(f"{name}.nc_set_variable({nc!r})")

        coordinates = self.coordinates()
        if coordinates:
            out.append(f"{name}.set_coordinates({coordinates})")

        for term, value in self.datum.parameters().items():
            if isinstance(value, self._Data):
                value = value.creation_commands(
                    name=None,
                    namespace=namespace0,
                    indent=0,
                    string=False,
                    header=header,
                )
            else:
                value = repr(value)

            out.append(f"{name}.datum.set_parameter({term!r}, {value})")

        for term, value in self.coordinate_conversion.parameters().items():
            if isinstance(value, self._Data):
                value = value.creation_commands(
                    name=None,
                    namespace=namespace0,
                    indent=0,
                    string=False,
                    header=header,
                )
            else:
                value = repr(value)

            out.append(
                f"{name}.coordinate_conversion.set_parameter({term!r}, {value})"
            )

        domain_ancillaries = self.coordinate_conversion.domain_ancillaries()
        if domain_ancillaries:
            out.append(
                f"{name}.coordinate_conversion.set_domain_ancillaries({domain_ancillaries})"
            )

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    @_display_or_return
    def dump(
        self,
        display=True,
        _omit_properties=None,
        field=None,
        key="",
        _level=0,
        _title=None,
        _construct_names=None,
        _auxiliary_coordinates=None,
        _dimension_coordinates=None,
    ):
        """A full description of the coordinate reference construct.

        Returns a description of all properties, including those of
        components.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        indent0 = "    " * _level
        indent1 = "    " * (_level + 1)

        if _title is None:
            string = [
                f"{indent0}Coordinate Reference: {self.identity(default='')}"
            ]
        else:
            string = [indent0 + _title]

        # Coordinate conversion parameter-valued terms
        coordinate_conversion = self.get_coordinate_conversion()
        for term, value in sorted(coordinate_conversion.parameters().items()):
            string.append(f"{indent1}Coordinate conversion:{term} = {value}")

        # Coordinate conversion domain ancillary-valued terms
        if _construct_names:
            for term, key in sorted(
                coordinate_conversion.domain_ancillaries().items()
            ):
                if key in _construct_names:
                    construct_name = (
                        "Domain Ancillary: "
                        + _construct_names.get(key, f"key:{key}")
                    )
                else:
                    construct_name = ""

                string.append(
                    f"{indent1}Coordinate conversion:{term} = {construct_name}"
                )
        else:
            for term, value in sorted(
                coordinate_conversion.domain_ancillaries().items()
            ):
                string.append(
                    f"{indent1}Coordinate conversion:{term} = {value}"
                )

        # Datum parameter-valued terms
        datum = self.get_datum()
        for term, value in sorted(datum.parameters().items()):
            string.append(f"{indent1}Datum:{term} = {value}")

        # Coordinates
        if _construct_names:
            for key in sorted(self.coordinates(), reverse=True):
                coord_name = _construct_names.get(key, f"key:{key}")
                coord = f"{coord_name}"
                if key in _dimension_coordinates:
                    coord = "Dimension Coordinate: " + coord
                elif key in _auxiliary_coordinates:
                    coord = "Auxiliary Coordinate: " + coord

                string.append(f"{indent1}{coord}")
        else:
            for identifier in sorted(self.coordinates()):
                string.append(f"{indent1}Coordinate: {identifier}")

        return "\n".join(string)

    @_manage_log_level_via_verbosity
    def equals(
        self, other, rtol=None, atol=None, verbose=None, ignore_type=False
    ):
        """Whether two coordinate reference constructs are the same.

        Equality is strict by default. This means that:

        * the datum and coordinate conversion components must have the
          same string and numerical parameters.

        The dimension coordinate, auxiliary coordinate and domain
        ancillary constructs of the coordinate reference constructs are
        *not* considered, because they may only be correctly interpreted
        by the field constructs that contain the coordinate reference
        constructs in question. They are, however, taken into account when
        two fields constructs are tested for equality.

        {{equals tolerance}}

        Any type of object may be tested but, in general, equality is only
        possible with another coordinate reference construct, or a
        subclass of one. See the *ignore_type* parameter.

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

            {{ignore_type: `bool`, optional}}

        :Returns:

            `bool`
                Whether the two coordinate reference constructs are equal.

        **Examples**

        >>> c = {{package}}.{{class}}(
        ...     coordinates=['dimensioncoordinate0']
        ... )
        >>> c.equals(c)
        True
        >>> c.equals(c.copy())
        True
        >>> c.equals('not a coordinate reference')
        False

        """
        pp = super()._equals_preprocess(
            other, verbose=verbose, ignore_type=ignore_type
        )
        if pp is True or pp is False:
            return pp

        other = pp

        coords0 = self.coordinates()
        coords1 = other.coordinates()
        if len(coords0) != len(coords1):
            logger.info(
                f"{self.__class__.__name__}: Different sized collections of "
                f"coordinates ({coords0}, {coords1})"
            )

            return False

        if not self.coordinate_conversion.equals(
            other.coordinate_conversion,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_type=ignore_type,
        ):
            logger.info(
                f"{self.__class__.__name__}: Different coordinate conversions"
            )

            return False

        if not self.datum.equals(
            other.datum,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_type=ignore_type,
        ):
            logger.info(f"{self.__class__.__name__}: Different datums")

            return False

        # Still here? Then the two coordinate references are as equal
        # as can be ascertained in the absence of domains.
        return True

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The ``standard_name`` coordinate conversion parameter, preceded
          by ``'standard_name:'``.
        * The ``grid_mapping_name`` coordinate conversion parameter,
          preceded by ``'grid_mapping_name:'``.
        * The netCDF variable name (corresponding to a netCDF grid mapping
          variable), preceded by ``'ncvar%'``.
        * The value of the *default* parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('coordinatereference0')
        >>> c.identity()
        'standard_name:atmosphere_hybrid_height_coordinate'

        >>> c = f.get_construct('coordinatereference1')
        >>> c.identity()
        'grid_mapping_name:rotated_latitude_longitude'

        >>> c = {{package}}.{{class}}()
        >>> c.identity()
        ''
        >>> c.identity(default='no identity')
        'no identity'

        """
        for prop in ("standard_name", "grid_mapping_name"):
            n = self.coordinate_conversion.get_parameter(prop, None)
            if n is not None:
                return f"{prop}:{n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Returns all possible identities.

        The identities comprise:

        * The ``standard_name`` coordinate conversion parameter, preceded
          by ``'standard_name:'``.
        * The ``grid_mapping_name`` coordinate conversion parameter,
          preceded by ``'grid_mapping_name:'``.
        * The netCDF variable name (corresponding to a netCDF grid mapping
          variable), preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('coordinatereference0')
        >>> c.identities()
        ['standard_name:atmosphere_hybrid_height_coordinate']

        >>> c = f.get_construct('coordinatereference1')
        >>> c.identities()
        ['grid_mapping_name:rotated_latitude_longitude',
         'ncvar%rotated_latitude_longitude']
        >>> for i in c.identities(generator=True):
        ...     print(i)
        ...
        grid_mapping_name:rotated_latitude_longitude
        ncvar%rotated_latitude_longitude

        >>> c = {{package}}.{{class}}()
        >>> c.identities()
        []

        """
        g = self._iter(body=self._identities_iter(), **kwargs)
        if generator:
            return g

        return list(g)



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class LinearSubarrayTest(unittest.TestCase):
    """Unit test for the LinearSubarrayArray class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_LinearSubarray_compressed_dimensions(self):
        """Test `LinearSubarray.compressed_dimensions`."""
        cd = {1: (1,)}
        x = cfdm.LinearSubarray(compressed_dimensions=cd)
        self.assertEqual(x.compressed_dimensions(), cd)

        x = cfdm.LinearSubarray()
        with self.assertRaises(ValueError):
            x.compressed_dimensions()

    def test_LinearSubarray_get_filename(self):
        """Test LinearSubarray.get_filename."""
        x = cfdm.LinearSubarray(
            data=123, parameters={}, dependent_tie_points={}
        )
        with self.assertRaises(AttributeError):
            x.get_filename()


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import logging

from . import core, mixin
from .decorators import _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class DomainAxis(
    mixin.NetCDFDimension,
    mixin.NetCDFUnlimitedDimension,
    mixin.Container,
    core.DomainAxis,
):
    """A domain axis construct of the CF data model.

    A domain axis construct specifies the number of points along an
    independent axis of the domain. It comprises a positive integer
    representing the size of the axis. In CF-netCDF it is usually
    defined either by a netCDF dimension or by a scalar coordinate
    variable, which implies a domain axis of size one. The field
    construct's data array spans the domain axis constructs of the
    domain, with the optional exception of size one axes, because
    their presence makes no difference to the order of the elements.

    **NetCDF interface**

    The netCDF dimension name of the construct may be accessed with
    the `nc_set_dimension`, `nc_get_dimension`, `nc_del_dimension`,
    and `nc_has_dimension` methods.

    Whether or not the netCDF is unlimited may be accessed with the
    `nc_is_unlimited` and `nc_set_unlimited` methods.

    The netCDF dimension group structure may be accessed with the
    `nc_set_dimension`, `nc_get_dimension`, `nc_dimension_groups`,
    `nc_clear_dimension_groups`, and `nc_set_dimension_groups` methods.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, size=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            size: `int`, optional
                The size of the domain axis.

                The size may also be set after initialisation with the
                `set_size` method.

                *Parameter example:*
                  ``size=192``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(size=size, source=source, copy=copy)

        self._initialise_netcdf(source)

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        return f"size({self.get_size('')})"

    def _identities_iter(self):
        """Return all possible identities.

        See `identities` for details and examples.

        :Returns:

            generator
                The identities.

        """
        n = self.nc_get_dimension(None)
        if n is not None:
            yield f"ncdim%{n}"

    def creation_commands(
        self, namespace=None, indent=0, string=True, name="c", header=True
    ):
        """Returns the commands to create the domain axis construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Field.creation_commands`

        :Parameters:

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.DomainAxis(size=12)
        >>> x.nc_set_dimension('time')
        >>> print(x.creation_commands(header=False))
        c = {{package}}.DomainAxis()
        c.set_size(12)
        c.nc_set_dimension('time')

        """
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        out = []

        if header:
            out.append("#")
            out.append(f"# {self.construct_type}:")
            identity = self.identity()
            if identity:
                out[-1] += f" {identity}"

        out.append(f"{name} = {namespace}{self.__class__.__name__}()")

        size = self.get_size(None)
        if size is not None:
            out.append(f"{name}.set_size({size})")

        nc = self.nc_get_dimension(None)
        if nc is not None:
            out.append(f"{name}.nc_set_dimension({nc!r})")

        if self.nc_is_unlimited():
            out.append(f"{name}.nc_set_unlimited({True})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = (f"\n{indent}").join(out)

        return out

    @_manage_log_level_via_verbosity
    def equals(self, other, verbose=None, ignore_type=False):
        """Whether two domain axis constructs are the same.

        Equality is strict by default. This means that:

        * the axis sizes must be the same.

        Any type of object may be tested but, in general, equality is
        only possible with another domain axis construct, or a
        subclass of one. See the *ignore_type* parameter.

        NetCDF elements, such as netCDF variable and dimension names,
        do not constitute part of the CF data model and so are not
        checked.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{verbose: `int` or `str` or `None`, optional}}

            {{ignore_type: `bool`, optional}}

        :Returns:

            `bool`
                Whether the two domain axis constructs are equal.

        **Examples**

        >>> d.equals(d)
        True
        >>> d.equals(d.copy())
        True
        >>> d.equals('not a domain axis')
        False

        >>> d = {{package}}.DomainAxis(1)
        >>> e = {{package}}.DomainAxis(99)
        >>> d.equals(e, verbose=3)
        DomainAxis: Different axis sizes: 1 != 99
        False

        """
        pp = super()._equals_preprocess(
            other, verbose=verbose, ignore_type=ignore_type
        )
        if pp is True or pp is False:
            return pp

        other = pp

        # Check that each axis has the same size
        self_size = self.get_size(None)
        other_size = other.get_size(None)
        if not self_size == other_size:
            logger.info(
                f"{self.__class__.__name__}: Different axis sizes: "
                f"{self_size} != {other_size}"
            )
            return False

        return True

    def identity(self, default=""):
        """Return the canonical identity.

        The identity is the first found of the following:

        1. The netCDF dimension name, preceded by 'ncdim%'.
        2. The value of the default parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of
                the default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> d = {{package}}.DomainAxis(size=9)
        >>> d.nc_set_dimension('time')
        >>> d.identity()
        'ncdim%time'
        >>> d.identity(default='no identity')
        'ncdim%time'
        >>> d.nc_del_dimension()
        'time'
        >>> d.identity()
        ''
        >>> d.identity(default='no identity')
        'no identity'

        """
        n = self.nc_get_dimension(None)
        if n is not None:
            return f"ncdim%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The netCDF dimension name, preceded by 'ncdim%'.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> d = {{package}}.DomainAxis(size=9)
        >>> d.nc_set_dimension('time')
        >>> d.identities()
        ['ncdim%time']
        >>> d.nc_del_dimension()
        'time'
        >>> d.identities()
        []
        >>> for i in d.identities(generator=True):
        ...     print(i)
        ...

        """
        g = self._iter(body=self._identities_iter(), **kwargs)
        if generator:
            return g

        return list(g)



from .abstract import SubsampledSubarray
from .mixin import QuadraticInterpolation


class QuadraticSubarray(QuadraticInterpolation, SubsampledSubarray):
    """A subarray of an array compressed by subsamplng.

    A subarray describes a unique part of the uncompressed array.

    The compressed data is reconstituted by quadratic interpolation.

    See CF section 8.3 "Lossy Compression by Coordinate Subsampling"
    and appendix J "Coordinate Interpolation Methods".

    >>> import numpy as np
    >>> tie_points = np.array([15, 135, 225, 255, 345])
    >>> w = np.array([5, 10, 5])
    >>> coords = {{package}}.QuadraticSubarray(
    ...     tie_points=tie_points,
    ...     tp_indices=(slice(0, 2, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(5,),
    ...     subarea_indices=(slice(0, 1, None),),
    ...     first=(True,),
    ...     parameters={'w': w},
    ... )
    >>> print(coords[...])
    [15.0 48.75 80.0 108.75 135.0]
    >>> coords = {{package}}.QuadraticSubarray(
    ...     tie_points=tie_points,
    ...     tp_indices=(slice(1, 3, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(4,),
    ...     subarea_indices=(slice(1, 2, None),),
    ...     first=(False,),
    ...     parameters={'w': w},
    ... )
    >>> print(coords[...])
    [173.88888888888889 203.88888888888889 225.0]
    >>> coords = {{package}}.QuadraticSubarray(
    ...     tie_points=tie_points,
    ...     tp_indices=(slice(3, 5, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(4,),
    ...     subarea_indices=(slice(2, 3, None),),
    ...     first=(True,),
    ...     parameters={'w': w},
    ... )
    >>> print(coords[...])
    [255.0 289.44444444444446 319.44444444444446 345.0]


    **Cell boundaries**

    When the tie points array represents bounds tie points, then the
    *shape* parameter describes the uncompressed bounds shape.

    >>> bounds_tie_points = np.array([0, 150, 240, 240, 360])
    >>> bounds = {{package}}.QuadraticSubarray(
    ...     tie_points=bounds_tie_points,
    ...     tp_indices=(slice(0, 2, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(5, 2),
    ...     subarea_indices=(slice(0, 1, None),),
    ...     first=(True,),
    ...     parameters={'w': w},
    ... )
    >>> print(bounds[...])
    [[0.0 33.2]
     [33.2 64.8]
     [64.8 94.80000000000001]
     [94.80000000000001 123.2]
     [123.2 150.0]]
    >>> bounds = {{package}}.QuadraticSubarray(
    ...     tie_points=bounds_tie_points,
    ...     tp_indices=(slice(1, 3, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(3, 2),
    ...     subarea_indices=(slice(1, 2, None),),
    ...     first=(False,),
    ...     parameters={'w': w},
    ... )
    >>> print(bounds[...])
    [[150.0 188.88888888888889]
     [188.88888888888889 218.88888888888889]
     [218.88888888888889 240.0]]
    >>> bounds = {{package}}.QuadraticSubarray(
    ...     tie_points=bounds_tie_points,
    ...     tp_indices=(slice(3, 5, None),),
    ...     subsampled_dimensions=(0,),
    ...     shape=(4, 2),
    ...     subarea_indices=(slice(2, 3, None),),
    ...     first=(True,),
    ...     parameters={'w': w},
    ... )
    >>> print(bounds[...])
    [[240.0 273.75]
     [273.75 305.0]
     [305.0 333.75]
     [333.75 360.0]]

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed data as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        (d1,) = tuple(self.compressed_dimensions())

        u_ab = self._select_data()

        u = self._quadratic_interpolation(
            ua=self._select_location(u_ab, {d1: 0}),
            ub=self._select_location(u_ab, {d1: 1}),
            w=self._select_parameter("w"),
            d1=d1,
        )
        u = self._post_process(u)

        if indices is Ellipsis:
            return u

        return u[indices]
